#include "regexp.h"

/*
 * Substitute the matching substring
 * This subroutine returns the length of the resulting
 * string. It does not write the terminating \0.
 * dst == 0 can be used to get the length without
 * doing actual substitution.
 */
int regnsub(r, src, srclen, dst0)
	const regexp *r;
	register const char *src;
	int srclen;
	char *dst0;
{
	register char *dst = dst0;
	const char *srcend = src + srclen;
	register char c;
	register int len;

	if( r == 0 || r->magic != REGEXP_MAGIC ) {
		regerror("no regular expression");
		return -1;
	}

	/*
	 * If dst == 0 we need only the length of resulting
	 * string.
	 */
	if( dst == 0 ) {
		while( src < srcend ) {
			c = *src++;
			if( c == '\\' ) {
				if( src >= srcend ) {
					dst++;
					break;
				}
				c = *src++;
				if( '0' <= c && c <= '9' ) 
					c -= '0';
				else {
					dst += (c != '&' && c != '\\') + 1;
					continue;
				}
			} else if( c != '&' ) {
				dst++;
				continue;
			} else
				c = 0;
			if( r->startp[c] == 0 )
				continue;
			len = r->endp[c] - r->startp[c];
			if( len == 0 )
				continue;
			if( len < 0 ) {
				regerror("damaged match string");
				return -1;
			}
			dst += len;
		}
		return (int) dst;
	}

	/*
	 * Do actual substitution
	 */
	while( src < srcend ) {
		c = *src++;
		if( c == '\\' ) {
			if( src >= srcend ) {
				*dst++ = '\\';
				break;
			}
			c = *src++;
			if( '0' <= c && c <= '9' ) 
				c -= '0';
			else {
				if( c != '&' && c != '\\' )
					*dst++ = '\\';
				*dst++ = c;
				continue;
			}
		} else if( c != '&' ) {
			*dst++ = c;
			continue;
		} else
			c = 0;
		if( r->startp[c] == 0 )
			continue;
		len = r->endp[c] - r->startp[c];
		if( len == 0 )
			continue;
		if( len < 0 ) {
			regerror("damaged match string");
			return -1;
		}
		memcpy(dst, r->startp[c], len);
		dst += len;
	}
	return dst - dst0;
}
