/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: beeptail.c,v $
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.5  1998/03/20  19:02:40  alex
 * *** empty log message ***
 *
 * Revision 1.4  1995/12/11  14:56:22  tim
 * *** empty log message ***
 *
 * Revision 1.4  1995/12/11  14:56:22  tim
 * *** empty log message ***
 *
 * Revision 1.3  1995/08/04  14:25:02  tim
 * *** empty log message ***
 *
 * Revision 1.3  1995/08/04  14:25:02  tim
 * *** empty log message ***
 *
 * Revision 1.2  1993/08/11  19:40:10  avg
 * Added copyright.
 *
 * Revision 1.2  1993/08/11  19:40:10  avg
 * Added copyright.
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 */
#ifndef lint
static char rcsid[] = "$Id: beeptail.c,v 1.1.1.1 1999/07/19 18:38:31 alex Exp $";
#endif

/*
 * A simplified version of tail which beeps
 * every time it reads a line with the specified pattern
 * (if it is specified).
 * The default
 */

#include <stdio.h>
#include <fcntl.h>

#define INIT    register char *bp = instring;
#define GETC()  (*bp++)
#define PEEKC() (*bp)
#define UNGETC(c) bp--
#define RETURN(c) return (c);
#define ERROR(c) regerr(c)
#include <regexp.h>

main(ac, av)
	int ac;
	char **av;
{
	char line[1024];
	char expbuf[1024];
	int in;
	int len, l, ll;
	char *p, *index(), *lp;
	char c;
	char *re = expbuf;
	extern char *optarg;
	extern int optind;
	char *res[100];
	int nres = 0;
	long pos;

	/*
	 * Parse option
	 */
	for(;;) {
		switch( getopt(ac, av, "r:") ) {
		case -1:
			if( optind == ac ) {
				in = 0;
				break;
			} else if( optind == ac - 1 ) {
				in = open(av[ac-1], O_RDONLY);
				if( in < 0 ) {
					fprintf(stderr, "%s: cannot open %s\n", av[0], av[ac-1]);
					exit(1);
				}
				break;
			}
			/*FALL THRU*/

		default:
		usage:
			fprintf(stderr, "Usage: %s [-r regexp]... file\n", av[0]);
			exit(1);

		case 'r':
			if( nres >= sizeof res / sizeof res[0] ) {
				fprintf(stderr, "%s: too many regular expressions\n", av[0]);
				exit(1);
			}
			res[nres++] = re;
			re = compile(optarg, re, &expbuf[sizeof expbuf - 1], 0);
			continue;
		}
		break;
	}

	/*
	 * Collect whole lines
	 */
	len = 0;
	for(;;) {
		pos = lseek(in, 0l, 1);
		l = read(in, &line[len], sizeof line - len - 1);
		if( l == 0 ) {
			if( lseek(in, 0l, 2) != pos )
				lseek(in, 0l, 0);
			else
				sleep(1);
			continue;
		}
		ll = len;
		len += l;
		line[len] = 0;
		lp = line;
		if( p = index(&line[ll], '\n') ) {
		next:   *p = 0;
			if( nres == 0 ) {
			beep:   write(1, "\7", 1);
			} else {
				for( l = 0 ; l < nres ; l++ ) {
					if( step(line, res[l]) )
						goto beep;
				}
			}
			*p = '\n';
			ll = (p - lp) + 1;
			write(1, lp, ll);
			len -= ll;
			if( len ) {
				lp = p + 1;
				if( p = index(lp, '\n') )
					goto next;
				bcopy(lp, line, len);
			}
		} else {
			if( len == sizeof line - 1) {
				write(1, line, len);
				len = 0;
			}
		}
	}
	/*NOTREACHED*/
}

/*
 * Report an error in reg expr
 */
regerr(errno)
	int errno;
{
	char *p;

	switch( errno ) {
	default:
		fprintf(stderr, "Error %d in regular expression\n", errno);
		exit(1);

	case 11: p = "Range endpoint too large."; break;
	case 16: p = "Bad number."; break;
	case 25: p = "``\\ digit'' out of range."; break;
	case 36: p = "Illegal or missing delimiter."; break;
	case 41: p = "No remembered search string."; break;
	case 42: p = "\\( \\) imbalance."; break;
	case 43: p = "Too many \\(."; break;
	case 44: p = "More than 2 numbers given in \\{ \\}."; break;
	case 45: p = "} expected after \\."; break;
	case 46: p = "First number exceeds second in \\{ \\}."; break;
	case 49: p = "[] imbalance."; break;
	case 50: p = "Regular expression(s) too long.";
	}
	fprintf(stderr, "Error in regular expression: %s\n", p);
	exit(1);
}
