/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: log.c,v $
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.8  1999/05/11 15:24:01  alex
 * *** empty log message ***
 *
 * Revision 1.6  1996/01/10  15:32:19  tim
 * New log format (with errors/drpos)
 *
 * Revision 1.5  1996/01/05  12:10:53  tim
 * before adding drops
 *
 * Revision 1.4  1995/09/07  09:29:53  tim
 * ???
 *
 * Revision 1.4  1995/09/07  09:29:53  tim
 * ???
 *
 * Revision 1.3  1993/09/22  23:25:53  avg
 * Added router logging functions.
 *
 * Revision 1.3  1993/09/22  23:25:53  avg
 * Added router logging functions.
 *
 * Revision 1.2  1993/08/11  19:40:10  avg
 * Added copyright.
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 */
#ifndef lint
static char rcsid[] = "$Id: log.c,v 1.1.1.1 1999/07/19 18:38:31 alex Exp $";
#endif

#include "manifest.h"
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h> /*adbr */


/*
 * File name formats (indexed by file_change)
 */
static char *fnf[] = {
	"%s/%s",                /* NONE */
	"%s/%s/%04d.%02d.%02d", /* DAILY */
	"%s/%s/%04d.%02d.%02d", /* WEEKLY */
	"%s/%s/%04d.%02d",      /* MONTHLY */
	"%s/%s/%04d",           /* YEARLY */
};

u_long  lastnotify = 0;     /* Time the last notification was timestamped with */

/*
 * Calculate the starting time for a file
 */
u_long
bof_time(t0)
	u_long t0;
{
	struct tm *tm;
	u_long bof;

	if( file_change == NONE )
		return 0;
	tm = localtime(&t0);
	bof = t0;
	switch( file_change ) {
	case YEARLY:
		bof -= tm->tm_yday * 24*60*60;
		goto daily;

	case MONTHLY:
		bof -= (tm->tm_mday - 1) * 24*60*60;
		goto daily;

	case WEEKLY:
		bof -= tm->tm_wday * 24*60*60;

	case DAILY:
	daily:
		bof -= tm->tm_hour * 60*60 +
		       tm->tm_min * 60 + tm->tm_sec;
		/* To prevent winter/summer change problem, increase time up to 1h */
		if ( t0 - bof > 7200 ) bof += 3600;
	}
	return bof;
}

/*
 * Open/create a link log file
 */
int
link_file(g, i, new, tt)
	struct GATEWAY *g;
	struct IF *i;
	int *new;
	u_long tt;
{
	struct tm *tm;
	char name[512];
	char *ldn, *bdn;

	tt = bof_time(tt);
	tm = localtime(&tt);
	bdn = base_dir;
	ldn = i->i_logdir;
	if( ldn[0] == '/' ) {
		ldn++;
		bdn = "";
	}
	sprintf(name, fnf[file_change], bdn, ldn,
		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
	return olf(name, new);
}

/*
 * Open/create a gateway log file
 */
int
gw_file(g, new, tt)
	struct GATEWAY *g;
	int *new;
	u_long tt;
{
	struct tm *tm;
	char name[512];
	char *perfix="bgp";

	tt = bof_time(tt);
	tm = localtime(&tt);
	sprintf(name, fnf[file_change], base_dir, g->g_name,
		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
	return olf(name, new);
}

/*
 * Open/create bgp log file
 */
int
bgp_file(g,new,tt)
	struct GATEWAY *g;
	int *new;
	u_long tt;
{
	struct tm *tm;
	char name[512];
	char bname[512];
	char *prefix = "bgp_";
	

	tt = bof_time(tt);
	tm = localtime(&tt);
	strcpy(bname,prefix);
	strcat(bname,g->g_name);

	sprintf(name, fnf[file_change], base_dir, bname,
		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
	return olf(name, new);
}

/*
 * Open/create a log file
 * Writes 1 to new if a new file has been created.
 */
int
olf(name, new)
	char *name;
	int *new;
{
# ifndef __FreeBSD__
	extern char *sys_errlist[];
	extern int errno;
# endif
	extern char *rindex(), *index();
	char *q, *p, dummy[1];
	int i;

	/* Try to open an existing file */
	*new = 0;
	if( (i = open(name, O_WRONLY | O_APPEND, 0)) >= 0 )
		return i;

	/* Try to create the file */
	*new = 1;
	if( (i = open(name, O_WRONLY | O_CREAT | O_APPEND, 0644)) >= 0 )
		return i;

	/* No luck. Try to create directories */
	p = dummy;
	while( q = rindex(name, '/') ) {
		*p = '/';
		*q = 0;
		if( mkdir(name, 0755) == 0 ) {
			dprintf(("MKDIR(%s)\n", name));
			*q = '/';
			while( p = index(q+1, '/') ) {
				*p = 0;
				dprintf(("MKDIR(%s)\n", name));
				if( mkdir(name, 0755) < 0 )
					goto derr;
				q = p;
				*p = '/';
			}
			goto dok;
		}
		if( errno != ENOENT )
			break;
		p = q;
	}
derr:   fprintf(stderr, "%s: cannot create directory %s %s\n", 
			av0, name, sys_errlist[errno]);
	return -1;

dok:    if( (i = open(name, O_WRONLY | O_CREAT | O_APPEND, 0644)) >= 0 )
		return i;
	fprintf(stderr, "%s: cannot create file %s\n", av0, name);
	return -1;
}

/*
 * Compose timestamp from Unix time
 */
char *
timestamp(tt)
	u_long tt;
{
	static char b[40];
	struct tm *tm;
	extern int hex_timestamp;

	if( !hex_timestamp ) {
		tm = localtime(&tt);
		sprintf(b, "%02d.%02d.%02d %02d:%02d:%02d",
			tm->tm_year % 100, tm->tm_mon + 1, tm->tm_mday,
			tm->tm_hour, tm->tm_min, tm->tm_sec);
	} else
		sprintf(b, "%08x", tt);
	return b;
}

/*
 * Notify if excessive errors on a link detected
 */
void
error_notify(g, i)
	struct GATEWAY *g;
	struct IF *i;
{
	char line[256];
	u_long tt = time(0);

	lastnotify = tt;
	if (i->i_out.s_inerrors > err_threshold) {
		sprintf(line, "%s IN ERRORS (%.2f%%): %s(%s) %s\n",
			timestamp(tt), i->i_out.s_inerrors,
			g->g_sname, i->i_name, i->i_comment);
		write(notify, line, strlen(line));
	}
	if (i->i_out.s_outerrors > err_threshold) {
		sprintf(line, "%s OUT ERRORS (%.2f%%): %s(%s) %s\n",
			timestamp(tt), i->i_out.s_outerrors,
			g->g_sname, i->i_name, i->i_comment);
		write(notify, line, strlen(line));
	}
	if (i->i_out.s_indrops > drop_threshold) {
		sprintf(line, "%s IN DROPS (%.2f%%): %s(%s) %s\n",
			timestamp(tt), i->i_out.s_indrops,
			g->g_sname, i->i_name, i->i_comment);
		write(notify, line, strlen(line));
	}
	if (i->i_out.s_outdrops > drop_threshold) {
		sprintf(line, "%s OUT DROPS (%.2f%%): %s(%s) %s\n",
			timestamp(tt), i->i_out.s_outdrops,
			g->g_sname, i->i_name, i->i_comment);
		write(notify, line, strlen(line));
	}
}

/* if b == NULL only create new bgp file for gateway */

bgp_opstat(g, b)
	struct GATEWAY *g;
	struct BGP *b;
{

	int new, f;
	u_long tt, t0;
	char line[256];
	struct stat st;
	
	dprintf(("Do bgp_opstat  for %s \n", g->g_name));
	tt = t0 = time(0);
	f = bgp_file(g,&new,t0);
	if(b == NULL) 
	   close (f);          /* Only create new bgp file for gateway */
	
	if(f < 0  || b==NULL)
        	return;

	/*
	 * Finally, if the file is not new
	 * check for the last file mod. time
	 */
	if( !new ) {
		fstat(f, &st);
		if( tt < (u_long) st.st_mtime )
			tt = st.st_mtime;
	}

	/* Write into the file */
	sprintf(line,"%s %s %s %d %s\n", 
		timestamp(tt), b->bgp_name,b->bgp_ip,b->bgp_as, bgpstats[b->bgp_state]);
	write(f, line, strlen(line));
	close(f);

	if( f == -1 )
		close(f);
	/* Update the state file */
	net_state_changed();

	/* Write the notification */
	if( notify < 0 )
		return;
	if( tt < lastnotify )
		tt = lastnotify;
	lastnotify = tt;
	sprintf(line, "%s BGP %s: %s(%s) %s %d (%s)\n", 
		timestamp(tt),bgpstats[b->bgp_state],
		g->g_sname,b->bgp_name,b->bgp_ip, 
		b->bgp_as, prtime(b->bgp_time));
	write(notify, line, strlen(line));
}

/*
 * Report the link status change
 * f0 is -1 if the link file is not already opened and
 * file descriptor if it is.
 */
void
link_opstat(g, i, f0, t0)
	struct GATEWAY *g;
	struct IF *i;
	int f0;
	u_long t0;      /* the reference time */
{
	u_long tt, bof;
	int new, f;
	struct stat st;
	char line[256];
	char *p;
	static char *opstats[] = {
		"?", "UP", "DOWN", "TESTING","STAT","INAC","SHUTDOWN","INACTIVE"
	};
# define OPSTATS(s) (s >= 0 && s < (sizeof opstats)/(sizeof (opstats[0]))? opstats[s]:opstats[0])
	dprintf(("Do link_opstat  for %s %s\n", g->g_name,i->i_name));
	new = 1;
	f = f0;
	if( f0 == -1 ) {
		f = link_file(g, i, &new, t0);
		if( f < 0 )
			return;
	}

	tt = i->i_cs.l_reftime - (g->g_uptime - i->i_cs.l_LastChange) / 100;
		
	if( tt < i->i_lastmsg )
		tt = i->i_lastmsg;      /* Keep records in chronological order */

	/* Do not use date prior to the beginning of the new file */
	bof = bof_time(t0);
	if( tt < bof )
		tt = bof;

	/*
	 * Finally, if the file is not new
	 * check for the last file mod. time
	 */
	if( !new ) {
		fstat(f, &st);
		if( tt < (u_long) st.st_mtime )
			tt = st.st_mtime;
	}

	/* Write into the file */
	sprintf(line, "%s %s\n", timestamp(tt),OPSTATS(i->i_cs.l_OperStatus));
	write(f, line, strlen(line));
	if( f0 == -1 )
		close(f);
	i->i_lastmsg = tt;
	
	/* Update the state file */
	net_state_changed();

	/* Write the notification */
	if( notify < 0 || i->i_ps.l_timestamp == 0 )
		return;
	if( tt < lastnotify )
		tt = lastnotify;
	lastnotify = tt;
	sprintf(line, "%s LINK %s: %s(%s) %s\n", timestamp(tt),
		OPSTATS(i->i_cs.l_OperStatus),
		g->g_sname, i->i_name, i->i_comment);
	write(notify, line, strlen(line));
}


/*
 * Write the link stat record to the file
 */
void
write_lstat(g, i)
	struct GATEWAY *g;
	struct IF *i;
{
	int f, new;
	char line[500];
	u_long t0 = i->i_out.s_time;
	static u_long LastTimeWrite=0;

	if( t0 == 0 )
		return;
	if( t0 < i->i_lastmsg )
		t0 = i->i_lastmsg;
	f = link_file(g, i, &new, t0);
	if( f < 0 )
		return;

	/* Put the status record at the beginning of the new file or expire time */
	if( new || time(0)-LastTimeWrite >= TIMEDUMP) {
		link_opstat(g, i, f, t0);
		LastTimeWrite=time(0);
	}


	if( i->i_cs.l_OperStatus != IF_UP ) {
		close(f);
		return;
	}

	/*                speed  inutil   inpkts     oututil   outpkt     inerr      outerr     indrop    outdrop */
	sprintf(line, "%s %.1f %.1f/%.1f %.1f/%.1f, %.1f/%.1f %.1f/%.1f, %.2f/%.2f  %.2f/%.2f, %.2f/%.2f %.2f/%.2f\n",
		timestamp(t0), i->i_cs.l_Speed / 1000.0,
		i->i_out.s_inutil, i->i_out.s_inpeakutil,
		i->i_out.s_inpckts, i->i_out.s_inpeakpckts,
		i->i_out.s_oututil, i->i_out.s_outpeakutil,
		i->i_out.s_outpckts, i->i_out.s_outpeakpckts,
		i->i_out.s_inerrors, i->i_out.s_inpeakerrors,
		i->i_out.s_outerrors, i->i_out.s_outpeakerrors,
		i->i_out.s_indrops, i->i_out.s_inpeakdrops,
		i->i_out.s_outdrops, i->i_out.s_outpeakdrops);
	write(f, line, strlen(line));
	close(f);
	i->i_lastmsg = t0;
}

/*
 * Report the gateway status change
 * (nf == 0 if no notification is required)
 */
void
gw_opstat(g, nf)
	struct GATEWAY *g;
{
	int f;
	int new;
	char line[256];
	u_long tt, t0, bof;
	char *s;
	struct stat st;
	static char *gstats[] = {
		"DOWN", "UP", "INAC", "FIRST"
	};


	dprintf(("Do gw_opstat  for %s\n", g->g_name));
	tt = t0 = time(0);
	f = gw_file(g, &new, t0);
	if( f < 0 )
		return;
	g->g_bof = bof = bof_time(t0);
	if( g->g_alive == G_UP ) {
		tt = time(0) - g->g_uptime / 100;
		if( tt < g->g_lastmsg )
			tt = g->g_lastmsg;  /* keep it in chronological order */

		/* Do not use date prior to the beginning of the new file */
		if( tt < bof )
			tt = bof;

		/*
		 * Finally, if the file is not new
		 * check for the last file mod. time
		 */
		if( !new ) {
			fstat(f, &st);
			if( tt < (u_long) st.st_mtime )
				tt = st.st_mtime;
		}
	} else          /* coming down */
		tt = time(0);

	if( g->g_alive == G_UP && nf != 2)
	   sprintf(line, "%s %s %ld\n", timestamp(tt), gstats[g->g_alive], g->g_uptime/100 );
        else 
	   sprintf(line, "%s %s\n", timestamp(tt), gstats[g->g_alive]);

	write(f, line, strlen(line));
	close(f);
	g->g_lastmsg = tt;

	/* Update the state file */
	net_state_changed();

	/* Write the notification */
	if( nf || notify < 0 )
		return;
	if( tt < lastnotify )
		tt = lastnotify;
	lastnotify = tt;
	sprintf(line, "%s ROUTER %s: %s\n", timestamp(tt),
		gstats[g->g_alive], g->g_sname);
	write(notify, line, strlen(line));
}

/*
 * Write the router stat record to the file
 */
void
write_gwstat(g)
	struct GATEWAY *g;
{
	int f;
	int new;
	char line[256];
        char mline[256];
	u_long tt, t0, bof;
	struct MEMPOLL *m;

	if( g->g_alive != G_UP || g->g_stat.gs_cnt == 0 )
		return;
	t0 = time(0);
	f = gw_file(g, &new, t0);
	if( f < 0 )
		return;
	if( new ) {
		g->g_bof = bof = bof_time(t0);
		tt = t0 - g->g_uptime / 100;
		if( tt < bof )
			tt = bof;
		sprintf(line, "%s UP\n", timestamp(tt));
		write(f, line, strlen(line));
	}

	sprintf(line, "%s %u/%u %lu/%lu %d", timestamp(t0),
		g->g_stat.gs_cpuavg / g->g_stat.gs_cnt,
		g->g_stat.gs_cpumax,
		g->g_stat.gs_memavg / g->g_stat.gs_cnt,
		g->g_stat.gs_memmin, g->g_stat.gs_temp);

        for(m = g->g_mem; m; m = m->mem_next) {
	   if (m->mem_valid != MEM_VALID_OK) {
	      /* sprintf(mline, " m%s:/E",m->mem_name);
		 strcat(line,mline);   */
              continue;
           }
	   if (m->mem_cnt != 0) {
	   sprintf(mline," M%s,%lu/%lu,%lu/%lu,%lu/%lu",m->mem_name,
		  m->mem_avgfree / m->mem_cnt,
                  m->mem_minfree,
                  m->mem_avgused / m->mem_cnt,
                  m->mem_maxused,
		  m->mem_avgmaxseg / m->mem_cnt,
                  m->mem_minmaxseg);

           strcat(line,mline);

           m->mem_cnt       = 0;
           m->mem_minfree   = 0;
           m->mem_avgfree   = 0;
           m->mem_maxused   = 0;
           m->mem_avgused   = 0;
	   m->mem_minmaxseg = 0;
           m->mem_avgmaxseg = 0;
           }
        }
        sprintf(mline,"\n");
        strcat(line,mline);
	g->g_stat.gs_cnt = 0;
	g->g_stat.gs_cpuavg = 0;
	g->g_stat.gs_cpumax = 0;
	g->g_stat.gs_memavg = 0;
	g->g_stat.gs_memmin = 1000000000;       /* some huge number */
	g->g_stat.gs_temp = 0;

	write(f, line, strlen(line));
	close(f);
	g->g_lastmsg = t0;
}
