/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: manifest.h,v $
 * Revision 1.2  1999/08/12 19:42:00  alex
 * Merged revision R-1-2-m
 *
 * Revision 1.1.1.1.2.1  1999/08/12 18:12:52  alex
 * Fist step of Sun5/FreeBSD merging
 *
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.5  1998/03/20  19:02:40  alex
 * RELCOM
 *
 * Revision 1.4  1995/09/07  09:29:53  tim
 * addede some fields in router and line status
 *
 * Revision 1.3  1995/08/04  14:25:02  tim
 * Added some output (avg line load etc.)
 *
 * Revision 1.2  1993/08/11  19:40:10  avg
 * Added copyright.
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <sys/cdefs.h>
#include <library/asn1.h>
#include "config.h"

#define DEBUG

#ifdef DEBUG
#define dprintf(x)      { if( debug ) printf x ; }
#else
#define dprintf(x)
#endif

/* Line configuration flags */
/* Check inactive input */
# define IC_NONE     00 /* No checks */
# define IC_INACTIVE 01 /* Check inactive input */
# define IC_INACTIVE_ATM 02 /* Check inactive input in case of AAL5 type only */

/*
 * Line status parameters
 */
struct LSTAT {
	u_long  l_timestamp;    /* the gateway's timestamp */
	u_long  l_reftime;      /* reference Unix time */
	/* Last change of list status.Cisco has bug about LastChange
	 * if interface is DOWN.
	 */ 
	u_long	l_lastchtt;
	/*
	 * SNMP variables
	 */
	u_long  l_Speed;        /* line speed, bps */
	int     l_OperStatus;   /* operational status */
	long    l_LastChange;   /* time of last change of op stat */
	u_long  l_InOctets;     /* # of input bytes */
	u_long  l_InUcastPkts;  /* # of input unicast packets */
	u_long  l_InNUcastPkts; /* # of input multicast packets */
	u_long  l_InDrops;      /* # of discarded input packets */
	u_long  l_InErrors;     /* # of input erorrs */
	u_long  l_OutOctets;    /* # of output bytes */
	u_long  l_OutUcastPkts; /* # of output unicast packets */
	u_long  l_OutNUcastPkts;/* # of output multicast packets */
	u_long  l_OutDrops;     /* # of discarded output packets */
	u_long  l_OutErrors;    /* # of output errors */
};

/*
 * Line aggregation counters
 */
struct LAGG {
	/* Input stats */
	double  s_inutil;       /* aggregated in utilization */
	double  s_inpeakutil;   /* peak in utilization */
	double  s_inpckts;      /* aggragated in pckts per second */
	double  s_inpeakpckts;  /* peak in pckts per second */

	/* Output stats */
	double  s_oututil;      /* aggregated out utilization */
	double  s_outpeakutil;  /* peak out utilization */
	double  s_outpckts;     /* aggragated out pckts per second */
	double  s_outpeakpckts; /* peak out pckts per second */

	/* Error stats */
	double  s_inerrors;     /* input errors / packet %% */
	double  s_inpeakerrors; /* peak input errors */
	double  s_outerrors;    /* output errors / packet %% */
	double  s_outpeakerrors;/* peak output errors */
	double  s_indrops;      /* input drops / packet %% */
	double  s_inpeakdrops;  /* peak input drops */
	double  s_outdrops;     /* output drops / packet %% */
	double  s_outpeakdrops; /* peak output drops */

	/* Timestamp */
	u_long  s_time;         /* timestamp */
};

/*
 * Gateway aggregation counters
 */
struct GAGG {
	unsigned gs_cnt;        /* # of measurements */
	unsigned gs_cpuavg;     /* Average CPU usage */
	unsigned gs_cpumax;     /* Maximal CPU usage */
	u_long   gs_memavg;     /* Average free memory */
	u_long   gs_memmin;     /* Minimal free memory */
	int      gs_temp;       /* Temperature of air */
};

/*
 * The time scheduler node type
 */
struct SNMP_ACTION {
	struct SNMP_ACTION *a_next;  /* Next action */
	u_long        a_time;  /* Alarm time */
	void         (*a_f)();  /* Function to call at the time */
};

/*
 * Polling state automaton states
 */
enum POLLSTAT {
	READ_IFNUM,     /* Obtaining the number of interfaces */
	READ_ENVDESCR,  /* Read the description of cisco environmantal card */
	READ_IFDESCR,   /* Reading interface description (MAP) */
	READ_IFINDEX,   /* Link via ifIndex(MAP) */
	READ_IFTYPE,	/* Get type of interface */
	READ_BGP,       /* Obtaining the bgp indet for BGP ALL */
	POLL_IF,        /* Polling interface */
	POLL_UP,	/* Polling gateway if it don't response us */
	READ_MEM,	/* Reading mem polls */
};

/* Struct for link Ifindex <-> Num interface */
struct MAP {
  char *m_val;
  int   m_key;
};

struct IF {
    /* Configuration data */
    char  *i_name;              /* Name of the interface */
    char  *i_logdir;            /* Log directory name */
    char  *i_comment;           /* Comment about the link */
    float i_maxinutil;          /* Max input line utilization */
    float i_maxoututil;         /* Max output line utilization */
    int   i_type;               /* Interface type */

    /* Polling state */
    int    i_index;             /* GW-specific interface index */
    int    i_reqid;             /* Current request id */
    u_long i_lastmsg;           /* Time the last message was logged */

    /* Line status */
    struct LSTAT i_cs;          /* Current link state */
    struct LSTAT i_ps;          /* Previous link state */
    u_long  i_overloadtime;     /* overload start time or 0 */
    u_long  i_errorstime;       /* high error rate start time or 0 */
    u_long  i_dropstime;        /* high drops start time or 0 */
    /* Statistics accumulators */
    struct LAGG  i_acc;         /* accumulated parameters */
    struct LAGG  i_out;         /* snapshot before printing out */
    struct MAP	 i_map;		/* map to ifIndex */
    char    i_icmode;		/* embedded check mode */
};

/*
 * The internal representation of the configuration file
 */
struct GATEWAY {
	struct SNMP_ACTION   g_act;          /* Polling action */
	struct SNMP_ACTION   g_updact;       /* Stat updating */

	/* Configuration data */
	struct GATEWAY *g_next;         /* Next entry */
	char           *g_name;         /* Hostname */
	char           *g_sname;        /* Short hostname */
	char           *g_cname;        /* Community name */
	int             g_poll;         /* Polling period (0 if default) */
	int             g_n;            /* Gateway # (to calculate time skew) */
	int             g_nwait;        /* # of link responces to wait for */

	/* Polling state */
	struct snmp_session *g_sess;    /* SNMP session data */
	enum POLLSTAT   g_ps;           /* Polling state automaton */
	char            g_alive;        /* Gateway is alive */
	char            g_cisco;        /* Is GW a cisco router? */
	char            g_ciscoenv;     /* Is cisco env monitor installed? */
	char            g_unused;
	u_long          g_uptime;       /* Gateway uptime (ticks) */
	int             g_inac;         /* Count of full lossing POLL_IFes */
	int             g_ngwifs;       /* Number of GW's hardware ifs */
	int             g_cin;          /* Current interface number */
	int             g_reqid;        /* Current request id */
	u_long          g_lastreq;      /* Time the last req was sent */
	u_long          g_lastmsg;      /* Time the last msg was logged */
	u_long          g_bof;          /* Current log file starting time */

	/* Router statistics (cisco only) */
	unsigned        g_cpubusy;      /* current cpu usage */
	u_long          g_freemem;      /* current free memory in K */
	u_long          g_memlowtime;   /* Memory low start time or 0 */
	u_long          g_cpubusytime;  /* CPU too busy start time or 0 */
	struct GAGG     g_stat;
	u_long		g_lastchst;	/* Time last change state of cisco */

	/* Mem polls */
	struct MEMPOLL	*g_mem;		/* Point to mem poll */

	/* BGP */
	struct BGP      *g_bgp;         /* BGP entry */
	oid             g_oidnext[56];  /* Tmp buffer */
	int             g_oidlen;
	int 		g_bgpall;	/* Poll all bgp peers */

	/* Interfaces */
	int             g_nifs, g_nifs_polled;   /* Number of polled interfaces */
	struct IF       g_ifs[1];       /* Flexible array of interfaces */
	
};

/*
 * Log file change periods
 */
enum FILE_CHANGE {
	NONE,           /* do not change log file */
	DAILY,          /* change log file daily */
	WEEKLY,         /* change log file weekly */
	MONTHLY,        /* change log file monthly */
	YEARLY          /* change log file yearly */
};

/*
* Receive state
*/

enum RECV_INDEX {
	RECV_GAT,
	RECV_INT,
	RECV_BGP,
	RECV_MEM
};

/*
 * BGP neighbor
 */
struct BGP {
	struct BGP *bgp_next;          /* Next Entry */
	char   *bgp_name;              /* Short name */
	char   *bgp_dname;             /* DNS name of peer */
	char   *bgp_ip;                /* Local ip address */
	int    bgp_as;                 /* AS number */
	int    bgp_state;              /* Bgp State */
	int    bgp_valid;	       /* Is there this bgp peer on router */
	int    bgp_reqid;              /* Current request id */
	u_long bgp_time;       	       /* Time of current state */
};

#define BGP_VALID_BAD	0
#define BGP_VALID_OK	1

static char *bgpstats[] = {
                "?", "IDLE", "CONNECT",
                "ACTIVE", "OPENSENT", "OPENCONF",
                "ESTABLISHED", "INAC"
};

/* 
* MEMPOLL 
*/

struct MEMPOLL {
	struct MEMPOLL *mem_next;	/* Next Entry */
	int		mem_type;	/* Type of memory */
	int		mem_valid;	/* Is there mem poll  on router */
        char           *mem_name;       /* Name of memory */
	u_long		mem_free;	/* Free memory */
	u_long		mem_used;	/* Used memory */	
	u_long		mem_maxseg;	/* Size of max buf */
	u_long          mem_memlowtime; /* Memory low start time or 0 */
	u_long		mem_minmem;	/* Min memory for this poll */
	u_long		mem_minfree;	/* Statistic for min free memory */
	u_long		mem_avgfree;	/* Statistic for avg free memory */
	u_long          mem_maxused;    /* Statistic for max used memory */
        u_long          mem_avgused;    /* Statistic for avg used memory */
        u_long          mem_minmaxseg;  /* Statistic for min maxseg */
	u_long          mem_avgmaxseg;  /* Statistic for avg maxseg */
        unsigned        mem_cnt;        /* # of measurements */ 
	int    		mem_reqid;	/* Current request id */

};

#define MEM_VALID_BAD	0
#define MEM_VALID_OK	1

static char *mempolls[] = {
	"?","PROCESSOR", "I/O", "PCI" , "FAST", "MULTIBUS"
};

static char *ifmempolls[] = {
	"?", "P", "I", "V", "F", "U"
};

/* Global variables */
extern char *av0;               /* Program name (for diagnostics) */
extern struct GATEWAY *gwlist;  /* List of polled gateways */
extern int poll_period;         /* Polling period */
extern int aggr_period;         /* Aggregation period */
extern char base_dir[];         /* The base directory name */
extern char notify_file[];      /* File to store notifications in */
extern char state_file[];       /* File with the current network state */
extern char ifsum_file[];       /* File with the dump all network interfaces */
extern char debug_file[];       /* File for debug purpose */
extern char lock_file[];        /* File to lock and store proc number */
extern int ngateways;           /* The number of gateways */
extern int debug;
extern enum FILE_CHANGE file_change;    /* File change period */
extern int fulldomain;          /* Preserve the full domain name in Stats/Notify */

extern int notify;              /* File descriptor to write notifications to */
extern float err_threshold;     /* Max % of errors on the link */
extern float drop_threshold;    /* Max % of discards on the link */
extern int gw_maxtemp;          /* Max router temperature */
extern int gw_maxcpu;           /* Max router CPU utilization */
extern int gw_minmem;           /* Minimal free memory */
extern FILE *fifstat; /* File for the status table */


/* Library functions */
extern char *index();
extern char *rindex();
extern long time();

/* Local functions */
extern u_long tsharp();
extern char *prtime();

/* Start time for snmpstatd */
extern u_long StartTime;
extern int timeinstable;

/* Time of instable state of object */
#define TIME_INSTABLE 120

/* MKTIME */
#ifdef MKTIME
#define timelocal(x) (mktime(x))
#endif

# ifdef solaris2
#define   LOCK_SH   1    /* shared lock */
#define   LOCK_EX   2    /* exclusive lock */
#define   LOCK_NB   4    /* don't block when locking */
#define   LOCK_UN   8    /* unlock */
#define   LOCK_FCNTL
# endif
