/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: netstate.c,v $
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.8  1999/05/11 15:24:01  alex
 * *** empty log message ***
 *
 * Revision 1.7  1998/03/20  19:02:40  alex
 * RELCOM
 *
 * Revision 1.6  1996/01/10  15:32:19  tim
 * notification on drops
 *
 * Revision 1.5  1995/09/07  09:29:53  tim
 * output of overhead and errors
 *
 * Revision 1.5  1995/09/07  09:29:53  tim
 * output of overhead and errors
 *
 * Revision 1.4  1995/08/04  14:25:02  tim
 * Added some output (avg line load etc.)
 *
 * Revision 1.3  1993/09/22  23:25:53  avg
 * Print short router names.
 *
 * Revision 1.2  1993/08/11  19:40:10  avg
 * Added copyright.
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 */
#ifndef lint
static char rcsid[] = "$Id: netstate.c,v 1.1.1.1 1999/07/19 18:38:31 alex Exp $";
#endif

#include "manifest.h"
extern int gw_maxcpu;
extern int gw_minmem;

static update_scheduled;
extern FILE *stf;

/*static char *prtime();
*/


/*
 * Trigger the net state update
 */
void
net_state_changed()
{
	static struct SNMP_ACTION a;
	void net_state();
	long tt;

	if( update_scheduled || stf == 0 )
		return;
	if( time(0) - StartTime < (aggr_period/2)  )
		tt= (aggr_period/2);
	else 
		tt=2;

	update_scheduled = 1;
	a.a_f = net_state;
	a.a_time = time(0) + tt;
	timeout(&a);
}

/*
 * Write current network state to a file
 */
/*ARGSUSED*/
void
net_state(a)
	struct SNMP_ACTION *a;
{
	struct GATEWAY *g;
	struct IF *i;
	struct BGP *b;
	int     ok;
	u_long  tt, curr_tt;
	char    lnickname[200], *cp;
	int     l;
	struct MEMPOLL *m;

	curr_tt = time(0);
	/* Check gateways */
	ok = 1;
	fseek(stf, 0l, 0);
	for( g = gwlist ; g ; g = g->g_next ) {
		if( g->g_alive == 1) {
			if (g->g_cisco) {
				if (g->g_cpubusy > gw_maxcpu)
					fprintf(stf, "ROUTER %s cpu load %d%% (%s)\n",
						      g->g_sname, g->g_cpubusy, prtime (curr_tt - g->g_cpubusytime));
				if (gw_minmem > 0 && g->g_freemem < gw_minmem && g->g_freemem > 0 )
					fprintf(stf, "ROUTER %s free memory %dK (%s)\n",
						      g->g_sname, g->g_freemem, prtime(curr_tt - g->g_memlowtime));
			}

			for (m= g->g_mem; m; m = m->mem_next) {
			  if (m->mem_valid != MEM_VALID_OK)
				continue; 
			  if (m->mem_minfree < m->mem_minmem && m->mem_minfree > 0)
                             fprintf(stf, "ROUTER %s free %s memory %dK (%s)\n",
				g->g_sname, mempolls[m->mem_type], m->mem_minfree,
				prtime(curr_tt - m->mem_memlowtime)); 
			}
			continue;
		}
		fprintf(stf, "ROUTER %s DOWN (%s)\n",
			      g->g_sname, prtime (curr_tt - g->g_lastmsg));
		ok = 0;
	}

	/* Check links */
	for( g = gwlist ; g ; g = g->g_next ) {
		if( g->g_alive != 1 )
			continue;
		for( i = g->g_ifs ; i < &g->g_ifs[g->g_nifs] ; i++ ) {
			if( i->i_index < 0 || i->i_cs.l_timestamp == 0)
				continue;
			if (i->i_comment[0]) {
				if (cp = index(i->i_comment, ' ')) {
					l = cp - i->i_comment;
					strncpy (lnickname, i->i_comment, l);
					lnickname[l] = '\0';
					cp++;
				}
				else
					strcpy (lnickname, i->i_comment);

			} else
				strcpy(lnickname, i->i_logdir);
			if(!cp)
				cp = "";

			if( i->i_cs.l_OperStatus == IF_UP ) {
				if( (i->i_maxinutil > 0.0 && i->i_out.s_inutil > i->i_maxinutil) ||
				    (i->i_maxoututil > 0.0 && i->i_out.s_oututil > i->i_maxoututil) )
				{
					if (!(i->i_overloadtime))
						i->i_overloadtime = i->i_cs.l_reftime;
					fprintf(stf, "LINK %s (%s) load %5.1f%%(in) %5.1f%%(out) (%s) %s\n",
						lnickname, g->g_sname,
						i->i_out.s_inutil, i->i_out.s_oututil,
						prtime(curr_tt - i->i_overloadtime), cp);
				} else
					i->i_overloadtime = 0;

				if ( (i->i_out.s_inerrors > err_threshold) ||
				     (i->i_out.s_outerrors > err_threshold) )
				{
					if (!(i->i_errorstime))
						i->i_errorstime = i->i_cs.l_reftime;
					fprintf(stf, "LINK %s (%s) errors %5.1f%%(in) %5.1f%%(out) (%s) %s\n",
						lnickname, g->g_sname,
						i->i_out.s_inerrors, i->i_out.s_outerrors,
						prtime(curr_tt - i->i_errorstime), cp);
				} else
					i->i_errorstime = 0;

				if ( (i->i_out.s_indrops > drop_threshold) ||
				     (i->i_out.s_outdrops > drop_threshold)  )
				{
					if (!(i->i_dropstime))
						i->i_dropstime = i->i_cs.l_reftime;
					fprintf(stf, "LINK %s (%s) drops %5.1f%%(in) %5.1f%%(out) (%s) %s\n",
						lnickname, g->g_sname,
						i->i_out.s_indrops, i->i_out.s_outdrops,
						prtime(curr_tt - i->i_dropstime), cp);
				} else
					i->i_dropstime = 0;
			}
			else {
				tt = i->i_cs.l_reftime - (g->g_uptime - i->i_cs.l_LastChange) / 100;
				if (tt > i->i_cs.l_lastchtt) 
					tt=i->i_cs.l_lastchtt; 
				fprintf(stf, "LINK %s (%s) %s (%s) %s\n",
					lnickname, g->g_sname,
					i->i_cs.l_OperStatus == IF_TESTING ? "TESTING" : "DOWN",
					prtime(curr_tt - tt), cp);
				ok = 0;
			}
		}
	}

	/* Check bgp */
	for( g = gwlist ; g ; g = g->g_next ) {
		if( g->g_alive != 1 )
			continue;
		for (b=g->g_bgp; b; b=b->bgp_next) {
			if(b->bgp_valid == BGP_VALID_BAD) continue;
			if(b->bgp_state != BGP_ESTABLISHED) {
				fprintf(stf,"BGP %s (%s) %s (%s)\n",
				b->bgp_name,g->g_sname, bgpstats[b->bgp_state],
				prtime(b->bgp_time));
				ok = 0;
			}
		}
	}

	/* If everything's OK that's it */
	if( ok )
		fprintf(stf, "ALL ROUTERS, ALL LINKS AND ALL BGP PEERS ARE UP\n");
	fflush(stf);
	ftruncate(fileno(stf), ftell(stf));
	write_ifsums();
	update_scheduled = 0;
        find_loss_gateways();
}


/*
 *  gateway       2 aggregate period
 */
find_loss_gateways(void) {
  struct GATEWAY *g;
  extern void poll_send();

  for( g = gwlist ; g ; g = g->g_next ) {
    if( g->g_lastreq + aggr_period*2 < time(0) ) { /*    gateway */
      g->g_act.a_time = time(0) + 2;
      g->g_act.a_f = poll_send;
      timeout((struct ACTIVE *)g);
    }
  }
}
 
/*
 * Print the human-readable period of time.
 */

char *prtime(t)
	u_long t;
{
	static char buf[128];
	char *bp = buf;
	int n;

#define M      60
#define H      (60*M)
#define D      (24*H)

	if( t >= D ) {
		n = t / D;
		t %= D;
		sprintf(bp, "%dd", n);
		bp += strlen(buf);
		t -= t % H;             /* get rid of minutes */
	}
	if( t >= H ) {
		if( bp != buf )
			*bp++ = ' ';
		n = t / H;
		t %= H;
		sprintf(bp, "%dh", n);
		bp += strlen(buf);
		t -= t % M;     /* get rid of seconds */
	}
	if( t >= M ) {
		if( bp != buf )
			*bp++ = ' ';
		n = t / M;
		t %= M;
		sprintf(bp, "%dm", n);
		bp += strlen(buf);
	}
	if( t ) {
		if( bp != buf )
			*bp++ = ' ';
		sprintf(bp, "%ds", t);
	}
	return buf;
}

