/*                                             
 * Copyright (C) 1993, US Sprint
 * 
 * All Rights Reserved
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of US Sprint not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THISP_STATE
 * SOFTWARE.
 * 
 * Author: Vadim Antonov
 * 
 * $Log: poll.c,v $
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.12  1999/06/08 10:38:22  alex
 * *** empty log message ***
 *
 * Revision 1.11  1998/03/25  19:46:04  alex
 * Fixed the bug around session.retries
 *
 * Revision 1.10  1998/03/23  15:42:46  alex
 * Re-linking of the interfaces added.
 * Revision 1.8  1996/01/10  15:32:19  tim Poll for all errors
 * and drops
 * 
 * Revision 1.7  1995/12/11  14:56:22  tim set interfaces down if timeout
 * 
 * Revision 1.6  1995/09/07  09:29:53  tim output of overhead and errors
 * 
 * Revision 1.5  1995/08/04  14:25:02  tim Added some output (avg line load
 * etc.)
 * 
 * Revision 1.4  1993/12/16  20:49:11  avg Version 3 cisco env controllrer is
 * ok.
 * 
 * Revision 1.3  1993/09/22  23:25:53  avg Added collection of (cisco specific)
 * router statistics (CPU util, memory, temperature).
 * 
 * Revision 1.2  1993/08/11  19:40:10  avg Added copyright.
 * 
 * Revision 1.2  1993/08/11  19:40:10  avg Added copyright.
 * 
 * Revision 1.1  1993/08/10  17:47:57  avg Initial revision
 * 
 * Revision 1.1  1993/08/10  17:47:57  avg Initial revision
 * 
 */
#ifndef lint
static char     rcsid[] = "$Id: poll.c,v 1.1.1.1 1999/07/19 18:38:31 alex Exp $";
#endif

#include "manifest.h"
#include <netinet/in.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <net-snmp/net-snmp-config.h>
#include <library/cmu_compat.h>
#include <library/snmp.h>
#include <library/snmp_impl.h>
#include <library/snmp_api.h>
#include <library/snmp_client.h>

long version;

/*
 * Variable names
 */
oid             v_sysUptime[] = {V_sysUptime};
oid             v_sysObjID[] = {V_sysObjID};
oid             v_ifNumber[] = {V_ifNumber};
oid             v_ifDescr[] = {V_ifDescr};
oid             v_ifType[] = {V_ifType};
oid             v_ifSpeed[] = {V_ifSpeed};
oid             v_ifOperStatus[] = {V_ifOperStatus};
oid             v_ifLastChange[] = {V_ifLastChange};
oid             v_ifInOctets[] = {V_ifInOctets};
oid             v_ifInUcastPkts[] = {V_ifInUcastPkts};
oid             v_ifInNUcastPkts[] = {V_ifInNUcastPkts};
oid             v_ifInDrops[] = {V_ifInDrops};
oid             v_ifInErrors[] = {V_ifInErrors};
oid             v_ifOutOctets[] = {V_ifOutOctets};
oid             v_ifOutUcastPkts[] = {V_ifOutUcastPkts};
oid             v_ifOutNUcastPkts[] = {V_ifOutNUcastPkts};
oid             v_ifOutDrops[] = {V_ifOutDrops};
oid             v_ifOutErrors[] = {V_ifOutErrors};

oid             v_ciscoProductID[] = {V_ciscoProductID};
oid             v_ciscoFreeMem[] = {V_ciscoFreeMem};
oid             v_ciscoAvgBusy[] = {V_ciscoAvgBusy};
oid             v_ciscoEnvPresent[] = {V_ciscoEnvPresent};
oid             v_ciscoEnvTemp[] = {V_ciscoEnvTemp};

oid             v_bgpVersion[] = {V_bgpVersion};
oid             v_bgpLocalAS[] = {V_bgpLocalAS};
oid             v_bgpPeerIdentifier[] = {V_bgpPeerIdentifier};
oid             v_bgpPeerState[] = {V_bgpPeerState};
oid             v_bgpPeerAdminStatus[] = {V_bgpPeerAdminStatus};
oid             v_bgpPeerNegotiatedVersion[] = {V_bgpPeerNegotiatedVersion};
oid             v_bgpPeerLocalAddr[] = {V_bgpPeerLocalAddr};
oid             v_bgpPeerLocalPort[] = {V_bgpPeerLocalPort};
oid             v_bgpPeerRemoteAddr[] = {V_bgpPeerRemoteAddr};
oid             v_bgpPeerRemotePort[] = {V_bgpPeerRemotePort};
oid             v_bgpPeerRemoteAs[] = {V_bgpPeerRemoteAs};
oid             v_bgpPeerInUpdates[] = {V_bgpPeerInUpdates};
oid             v_bgpPeerOutUpdates[] = {V_bgpPeerOutUpdates};
oid             v_bgpPeerInTotalMessages[] = {V_bgpPeerInTotalMessages};
oid             v_PeerOutTotalMessages[] = {V_PeerOutTotalMessages};
oid             v_bgpPeerLastError[] = {V_bgpPeerLastError};
oid             v_bgpPeerFsmEstablishedTime[] = {V_bgpPeerFsmEstablishedTime};

oid             v_ciscoMemoryPoolName[] = {V_ciscoMemoryPoolName};
oid             v_ciscoMemoryPoolAlternate[] = {V_ciscoMemoryPoolAlternate};
oid             v_ciscoMemoryPoolValid[] = {V_ciscoMemoryPoolValid};
oid             v_ciscoMemoryPoolUsed[] = {V_ciscoMemoryPoolUsed};
oid             v_ciscoMemoryPoolFree[] = {V_ciscoMemoryPoolFree};
oid             v_ciscoMemoryPoolLargestFree[] = {V_ciscoMemoryPoolLargestFree};

/* Length of the variable */
#define v_len(v)        (sizeof(v) / sizeof(oid))

/* Set the instance # */
#define v_instance(v, n)  v[v_len(v) - 1] = (n)


v_dump(v, sv)
	oid            *v;
{
	while (sv-- > 0)
		printf(".%d", *v++);
	printf("\n");
}

#define Snmp_add_null_var(p, n, l) \
	snmp_add_null_var(p, n, l)


int             poll_input();
void            poll_send();
void            poll_ptimo();
int 		check_uptime();

#define NEXTVAR var = var->next_variable
#define BIGINOFBGPID 10

int             MAP_STATE[] = {READ_IFDESCR, READ_IFINDEX};

/* Reset all requid */
void reset_reqid(g)
	struct GATEWAY *g;
{
	int             idx;
	struct IF      *gif;
	struct BGP     *b;
	struct MEMPOLL *m;
	
	g->g_reqid = 0;
	g->g_nwait = 0;
	for (idx = 0; idx < g->g_nifs; idx++) {
		gif = &(g->g_ifs[idx]);
		gif->i_reqid = 0;
	}
	for (b = g->g_bgp; b; b = b->bgp_next)
		b->bgp_reqid = 0;
	for (m = g->g_mem; m; m = m->mem_next)
		m->mem_reqid = 0;
}

/*
 * Reset gateway data to the initial state.
 */
void
reset_gwdata(g)
	struct GATEWAY *g;
{
	int             idx;
	struct IF      *gif;
	struct BGP     *b;
	struct MEMPOLL *m;

	g->g_ps = READ_IFNUM;
	g->g_alive = G_INAC;
        g->g_inac  = 0;
	g->g_reqid = 0;
	g->g_nwait = 0;
	g->g_uptime= 0;
	g->g_cisco = 0;
	g->g_lastchst = time(0);
	g->g_ciscoenv = 0;
	g->g_stat.gs_cnt = 0;
	g->g_stat.gs_cpuavg = 0;
	g->g_stat.gs_cpumax = 0;
	g->g_stat.gs_memavg = 0;
	g->g_stat.gs_memmin = 1000000000;	/* some huge number */
	g->g_stat.gs_temp = 0;
	g->g_oidlen = 0;
	for (idx = 0; idx < g->g_nifs; idx++) {
		gif = &(g->g_ifs[idx]);
		gif->i_index = -1;
		gif->i_reqid = 0;
		gif->i_type = 0;
		/* reset interface statistics */
		bzero(&(gif->i_ps), sizeof(struct LSTAT));
		bzero(&(gif->i_cs), sizeof(struct LSTAT));
		bzero(&(gif->i_acc), sizeof(struct LAGG));
		bzero(&(gif->i_out), sizeof(struct LAGG));
		/* we have to set up interface into inaccessable state */
		gif->i_cs.l_OperStatus = IF_INAC;
	}
	for (b = g->g_bgp; b; b = b->bgp_next) {
		b->bgp_reqid = 0;
		b->bgp_as = 0;
		b->bgp_valid = BGP_VALID_BAD;
		b->bgp_state = BGP_INAC;
		b->bgp_time = 0;
	}
	for (m = g->g_mem; m; m = m->mem_next) {
		m->mem_reqid     = 0;
                m->mem_cnt       = 0;
                m->mem_avgmaxseg = 0;
                m->mem_minmaxseg = 0;
                m->mem_avgused   = 0;
                m->mem_maxused   = 0;
                m->mem_avgfree   = 0;
                m->mem_minfree   = 0;
                m->mem_memlowtime= 0;
                m->mem_maxseg    = 0;
                m->mem_used      = 0;
                m->mem_free      = 0;
		m->mem_valid     = MEM_VALID_BAD;
	}
}


/*
 * Open session and launch the state machine for a gateway.
 */
void
start_poll(g)
	struct GATEWAY *g;
{
	struct snmp_session session, *ss;
	int             j;

	bzero(&session, sizeof session);
	session.peername = g->g_name;
	session.community = (u_char *) g->g_cname;
	session.community_len = strlen(g->g_cname);
	session.retries = SNMP_RETRIES;
	session.timeout = SNMP_DEFAULT_TIMEOUT;
	session.authenticator = NULL;
	session.callback = poll_input;
	session.callback_magic = (void *) g;
	session.version = version;

	ss = snmp_open(&session);
	if (ss == 0) {
		fprintf(stderr, "%s: cannot start SNMP session with %s\n",
			av0, g->g_name);
		exit(1);
	}
	g->g_sess = ss;
	reset_gwdata(g);
	g->g_alive = G_FIRST;
	poll_send((struct SNMP_ACTION *) g);
}

/*
 * Send the poll request
 */
void
poll_send(a)
	struct SNMP_ACTION  *a;
{
	struct GATEWAY *g = (struct GATEWAY *) a;
	struct IF      *ii;
	struct BGP     *b;
	struct MEMPOLL *m;
	int             i, j;
	struct snmp_pdu *pdu;

	dprintf(("poll_send(%s) state %d\n", g->g_name, g->g_ps));

	g->g_lastreq = time(0);
	/*
	 * Regular interface polling -- multiple simultaneous requests per
	 * gateway
	 */
	if (g->g_ps == POLL_IF) {
		for (i = 0; i < g->g_nifs; i++) {
			/* system.sysUptime is used as timestamp */
			ii = &g->g_ifs[i];
			if ((j = ii->i_index) < 0)
				continue;
			pdu = snmp_pdu_create(GET_REQ_MSG);
			Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));

			v_instance(v_ifSpeed, j);
			Snmp_add_null_var(pdu, v_ifSpeed, v_len(v_ifSpeed));
			v_instance(v_ifOperStatus, j);
			Snmp_add_null_var(pdu, v_ifOperStatus, v_len(v_ifOperStatus));
			v_instance(v_ifLastChange, j);
			Snmp_add_null_var(pdu, v_ifLastChange, v_len(v_ifLastChange));

			switch (ii->i_type) {

			case ifType_frame_relay:

				v_instance(v_ifInOctets, j);
				Snmp_add_null_var(pdu, v_ifInOctets, v_len(v_ifInOctets));
				v_instance(v_ifOutOctets, j);
				Snmp_add_null_var(pdu, v_ifOutOctets, v_len(v_ifOutOctets));
				break;
				
			case ifType_aal5:

				v_instance(v_ifInOctets, j);
				Snmp_add_null_var(pdu, v_ifInOctets, v_len(v_ifInOctets));
				v_instance(v_ifInUcastPkts, j);
				Snmp_add_null_var(pdu, v_ifInUcastPkts, v_len(v_ifInUcastPkts));
				v_instance(v_ifOutOctets, j);
				Snmp_add_null_var(pdu, v_ifOutOctets, v_len(v_ifOutOctets));
				v_instance(v_ifOutUcastPkts, j);
				Snmp_add_null_var(pdu, v_ifOutUcastPkts, v_len(v_ifOutUcastPkts));
				break;

			default:
				/* Ask all the interface data */
				v_instance(v_ifInOctets, j);
				Snmp_add_null_var(pdu, v_ifInOctets, v_len(v_ifInOctets));
				v_instance(v_ifInUcastPkts, j);
				Snmp_add_null_var(pdu, v_ifInUcastPkts, v_len(v_ifInUcastPkts));
				v_instance(v_ifInNUcastPkts, j);
				Snmp_add_null_var(pdu, v_ifInNUcastPkts, v_len(v_ifInNUcastPkts));
				v_instance(v_ifInDrops, j);
				Snmp_add_null_var(pdu, v_ifInDrops, v_len(v_ifInDrops));
				v_instance(v_ifInErrors, j);
				Snmp_add_null_var(pdu, v_ifInErrors, v_len(v_ifInErrors));

				v_instance(v_ifOutOctets, j);
				Snmp_add_null_var(pdu, v_ifOutOctets, v_len(v_ifOutOctets));
				v_instance(v_ifOutUcastPkts, j);
				Snmp_add_null_var(pdu, v_ifOutUcastPkts, v_len(v_ifOutUcastPkts));
				v_instance(v_ifOutNUcastPkts, j);
				Snmp_add_null_var(pdu, v_ifOutNUcastPkts, v_len(v_ifOutNUcastPkts));
				v_instance(v_ifOutDrops, j);
				Snmp_add_null_var(pdu, v_ifOutDrops, v_len(v_ifOutDrops));
				v_instance(v_ifOutErrors, j);
				Snmp_add_null_var(pdu, v_ifOutErrors, v_len(v_ifOutErrors));
				break;
			};

			if ( g->g_nwait++ == 0 )
			       g->g_inac++;

			if (!(ii->i_reqid = snmp_send(g->g_sess, pdu))) {
				fprintf(stderr, "%s: cant send if request to %s\n", av0, g->g_name);
				g->g_act.a_time = time(0) + 10;
				ii->i_reqid = -1;
				g->g_act.a_f = poll_ptimo;
				timeout((struct ACTIVE *) g);
				return;
			}
			dprintf(("IF send %s %s i_reqid %d g_nwait %d\n", 
				g->g_name, ii->i_name, ii->i_reqid,g->g_nwait));
		}

		for (m = g->g_mem; m; m = m->mem_next) {
			if (m->mem_valid != MEM_VALID_OK)
				continue;

			pdu = snmp_pdu_create(GET_REQ_MSG);
			Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
			v_instance(v_ciscoMemoryPoolUsed, m->mem_type);
			Snmp_add_null_var(pdu, v_ciscoMemoryPoolUsed, v_len(v_ciscoMemoryPoolUsed));
			v_instance(v_ciscoMemoryPoolFree, m->mem_type);
			Snmp_add_null_var(pdu, v_ciscoMemoryPoolFree, v_len(v_ciscoMemoryPoolFree));
			v_instance(v_ciscoMemoryPoolLargestFree, m->mem_type);
			Snmp_add_null_var(pdu,v_ciscoMemoryPoolLargestFree,v_len(v_ciscoMemoryPoolLargestFree));
			if ( g->g_nwait++ == 0 )
			       g->g_inac++ ;

			if (!(m->mem_reqid = snmp_send(g->g_sess, pdu))) {
				fprintf(stderr, "%s: cant send mem request to %s\n", av0, g->g_name);
				g->g_act.a_time = time(0) + 10;
				m->mem_reqid = -1;
				g->g_act.a_f = poll_ptimo;
				timeout((struct ACTIVE *) g);
				return;
			}
			dprintf(("MEM send %s %s mem_reqid %d\n", 
                          g->g_name, mempolls[m->mem_type],m->mem_reqid));
		}

		for (b = g->g_bgp; b; b = b->bgp_next) {
			register int    i;
			char            buff[16];
			register char  *so;
			long            lo;

			if (b->bgp_valid != BGP_VALID_OK)
				continue;

			pdu = snmp_pdu_create(GET_REQ_MSG);
			Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));

			bzero(&buff, sizeof(buff));
			i = strlen(b->bgp_ip);
			strncpy(buff, b->bgp_ip, i);
			i = BIGINOFBGPID;
			so = (char *) strtok(buff, ".");
			while (so) {
				lo = atol(so);
				v_bgpPeerState[i] = lo;
				v_bgpPeerRemoteAs[i] = lo;
				v_bgpPeerFsmEstablishedTime[i++] = lo;
				so = (char *) strtok(NULL, ".");
			}

			Snmp_add_null_var(pdu, v_bgpPeerState, v_len(v_bgpPeerState));
			Snmp_add_null_var(pdu, v_bgpPeerRemoteAs, v_len(v_bgpPeerRemoteAs));
			Snmp_add_null_var(pdu, v_bgpPeerFsmEstablishedTime, v_len(v_bgpPeerFsmEstablishedTime));
			if ( g->g_nwait++ == 0 )
				g->g_inac++;
			if (!(b->bgp_reqid = snmp_send(g->g_sess, pdu))) {
				fprintf(stderr, "%s: cant send bgp request to %s\n", av0, g->g_name);
				g->g_act.a_time = time(0) + 10;
				b->bgp_reqid = -1;
				g->g_act.a_f = poll_ptimo;
				timeout((struct ACTIVE *) g);
				return;
			}
			dprintf(("BGP send %s %s bgp_reqid %d\n", 
                          g->g_name, b->bgp_name, b->bgp_reqid));
		}
	}
	/*
	 * Single request per gateway
	 */
	switch (g->g_ps) {
	case READ_IFNUM:
		pdu = snmp_pdu_create(GET_REQ_MSG);
		Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		Snmp_add_null_var(pdu, v_ifNumber, v_len(v_ifNumber));
		Snmp_add_null_var(pdu, v_sysObjID, v_len(v_sysObjID));
		break;

	case READ_ENVDESCR:
		pdu = snmp_pdu_create(GET_REQ_MSG);
		Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		Snmp_add_null_var(pdu, v_ciscoEnvPresent, v_len(v_ciscoEnvPresent));
		break;

	case READ_BGP:
		pdu = snmp_pdu_create(GETNEXT_REQ_MSG);
		if (g->g_oidlen)
			Snmp_add_null_var(pdu, g->g_oidnext, g->g_oidlen);
		else
			Snmp_add_null_var(pdu, v_bgpPeerIdentifier, v_len(v_bgpPeerIdentifier));
		break;

	case READ_MEM:
		pdu = snmp_pdu_create(GET_REQ_MSG);
                Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		v_instance(v_ciscoMemoryPoolName, g->g_cin);
		Snmp_add_null_var(pdu,v_ciscoMemoryPoolName,v_len(v_ciscoMemoryPoolName));
		break;

	case POLL_UP:
		pdu = snmp_pdu_create(GET_REQ_MSG);
		Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		break;

	case READ_IFDESCR:
		pdu = snmp_pdu_create(GETNEXT_REQ_MSG);
		v_instance(v_ifDescr, g->g_cin);
		Snmp_add_null_var(pdu, v_ifDescr, v_len(v_ifDescr));
		break;

	case READ_IFINDEX:
		pdu = snmp_pdu_create(GET_REQ_MSG);
		Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		break;

	case READ_IFTYPE:
		pdu = snmp_pdu_create(GET_REQ_MSG);
		Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		for (i = g->g_nifs - 1; i >= 0; i--) {
			ii = &g->g_ifs[i];
			if ((j = ii->i_index) < 0)
				continue;
			v_instance(v_ifType, j);
			Snmp_add_null_var(pdu, v_ifType, v_len(v_ifType));
		}
		break;

	case POLL_IF:
		pdu = snmp_pdu_create(GET_REQ_MSG);
		Snmp_add_null_var(pdu, v_sysUptime, v_len(v_sysUptime));
		if ( g->g_nwait++ == 0 ) g->g_inac++;
		if (g->g_cisco) {
			Snmp_add_null_var(pdu, v_ciscoAvgBusy, v_len(v_ciscoAvgBusy));
			Snmp_add_null_var(pdu, v_ciscoFreeMem, v_len(v_ciscoFreeMem));
			if (g->g_ciscoenv)
				Snmp_add_null_var(pdu, v_ciscoEnvTemp, v_len(v_ciscoEnvTemp));
		}
	}
	if ( g->g_nwait ==  0 )
	       g->g_inac++;

	if (!(g->g_reqid = snmp_send(g->g_sess, pdu))) {
		fprintf(stderr, "%s: cant send to %s\n", av0, g->g_name);
		g->g_reqid = -1;
		g->g_act.a_time = time(0) + 10;
		g->g_act.a_f = poll_ptimo;
		timeout((struct ACTIVE *) g);
		return;
	}
}

/*
 * Pseudo-timeout
 */
void
poll_ptimo(a)
	struct SNMP_ACTION  *a;
{
	poll_input(NETSNMP_CALLBACK_OP_TIMED_OUT, (struct snmp_session *) 0, -1,
		   (struct snmp_pdu *) 0, (void *) a);

}

/*
 * Read the SNMP variable of integer type
 */
#define GETVAR(ls, name, t, err) { \
		if( var == 0 ) { \
			fprintf(stderr, "%s: %s got no in SNMP GET responce\n",\
			 av0, g->g_name); \
			err; \
		} \
		if( var->type != t ) { \
			fprintf(stderr, "%s: %s %s wrong SNMP name\n",\
			 av0, name, g->g_name ); \
			err; \
		} \
		ls = *var->val.integer; \
	}


/*
 * The SNMP input callback routine
 */
int
poll_input(op, s, reqid, pdu, magic)
	int             op;
	struct snmp_session *s;
	int             reqid;
	struct snmp_pdu *pdu;
	void           *magic;
{
	struct GATEWAY *g = (struct GATEWAY *) magic;
	struct IF      *i;
	struct BGP     *b;
	struct MEMPOLL *m;
	int             j, k, iff;
	struct variable_list *var;
	long            lo;
	struct LSTAT    l;
	struct BGP      nb;

        b = 0;
	i = 0;
	m = 0;

	/* Process dead gw */
	if (op == NETSNMP_CALLBACK_OP_TIMED_OUT) {
		if ( g->g_ps == POLL_IF ) {

		  /* check out if there is such requid for bgp, mem, gateway ??? 
                  * if not, suppouse it's very old request. Drop it.
                  * One more problem that i've got,request can be got out of order.
                  * I think it should be problem of callback sub into lib.
                  */
                  k = 1;
                  for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
  		  	if (reqid == i->i_reqid) {
			        dprintf(("GW %s HAS LOST IF %s g_nwait %d i_reqid %d \n", g->g_name,i->i_name,g->g_nwait,i->i_reqid));
				i->i_reqid=0;
				k=0;
                                break;
		  	}
		  }
	  	  for (b = g->g_bgp; b && k; b = b->bgp_next) {
		  	if (reqid == b->bgp_reqid) {
				dprintf(("GW %s HAS LOST BGP g_nwait %d bgp_reqid %d \n", g->g_name,g->g_nwait,b->bgp_reqid));
				b->bgp_reqid=0;
				k=0;
				break;
			}
		  }
		  for (m = g->g_mem; m && k; m = m->mem_next) {
                        if (reqid == m->mem_reqid) {
				dprintf(("GW %s HAS LOST MEM g_nwait %d mem_reqid %d\n", g->g_name,g->g_nwait,m->mem_reqid));
                        	m->mem_reqid=0;
                        	k=0;
				break;
			}
		  }

                  /* It would be better to continue. If we drop it and there is
                   * no more requests will come, the router may be lost.
                  */
                  if( k )
                     dprintf(("GW %s receive the packet without of reqid (nwait=%d)\n",g->g_name,g->g_nwait));

		}
		if( g->g_nwait > 0 ) g->g_nwait--;

		/* There is requests in the network yet. Wait a minute */
		if ( g->g_nwait > 0 ) return(1);
		/*
		* Count how many requests have been delayed and don't put
		*  router into INAC state until we reach limite of losses
		*/
		if ( g->g_inac < 2 ) {
		   reset_reqid(g);
		   g->g_act.a_time += g->g_poll ? g->g_poll : poll_period;
		   g->g_act.a_time += DEAD_POLL;
		   g->g_act.a_f = poll_send;
		   timeout((struct ACTIVE *) g);
		   return 1;
		} 
		
		if (g->g_alive == G_UP) {
			/*
			 * If gateway has experienced downtime its
			 * configuration has probably changed (it can change
			 * interface indexes).
			 */
			j = g->g_alive;
			/*
			 * Simulate all interfaces INAC state
			 */
			for (k = 0; k < g->g_nifs; k++) {
				i = &(g->g_ifs[k]);
				if (i->i_index >= 0 && i->i_cs.l_OperStatus == IF_UP) {
					i->i_cs.l_OperStatus = IF_INAC;
					link_opstat(g, i, -1, i->i_cs.l_reftime);
				}
			}
			/* reset_gwdata(g); */
			reset_reqid(g);
			g->g_alive = G_INAC;
			g->g_lastchst = time(0);
			gw_opstat(g, 0);
			for (b = g->g_bgp; b; b = b->bgp_next) {
				if (b->bgp_valid != BGP_VALID_OK ||
				    b->bgp_state == BGP_IDLE)
					continue;
				nb.bgp_state = BGP_INAC;
				/*bzero((char *) &nb, sizeof(nb));*/
				bgp_stat(g, b, &nb);
			}
			dprintf(("%s: GW IS INACCESSIBLE\n", g->g_name));
		}
		/* Keep pinging */
		g->g_ps = POLL_UP;	/* trying to catch gateway */
		g->g_act.a_time = g->g_lastreq + DEAD_POLL;
		g->g_act.a_f = poll_send;
		timeout((struct SNMP_ACTION *) g);
		return 1;
	}

	g->g_inac = 0; /* Drop the count of full lossing polls */

	if (op != NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE) {
		printf("REPLY: op != NETSNMP_CALLBACK_OP_RECEIVED MESSAGE (%d)\n", op);
		return 1;
	}
	/* Search for an interface OR gateway */
	if (reqid == 0 )
		return 0;

	b = 0;
	i = 0;
	m = 0;

	for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
		if (reqid == i->i_reqid) {
			if (g->g_ps != POLL_IF)
				return 0;
			i->i_reqid = 0;	/* Skip duplicates */
			iff = RECV_INT;
			goto if_found;
		}
	}

	for (b = g->g_bgp; b; b = b->bgp_next) {
		if (reqid == b->bgp_reqid) {
			if (g->g_ps != POLL_IF)
				return 0;
			b->bgp_reqid = 0;	/* Skip duplicates */
			iff = RECV_BGP;
			goto if_found;
		}
	}

	for (m = g->g_mem; m; m = m->mem_next) {
		if (reqid == m->mem_reqid) {
			if (g->g_ps != POLL_IF)
				return 0;
			m->mem_reqid = 0;	/* Skip duplicates */
			iff = RECV_MEM;
			goto if_found;
		}
	}

	iff = RECV_GAT;
	if (reqid != g->g_reqid)
		return 0;
	else
		g->g_reqid = 0;	/* Skip duplicates */
if_found:


	/*
	 * The only expected error is missing interface name
	 */
	if (pdu->errstat != SNMP_ERR_NOERROR) {
		if (pdu->errstat == SNMP_ERR_NOSUCHNAME) {
			switch (g->g_ps) {
			case READ_IFDESCR:
				dprintf(("%s: no if%d\n", g->g_name, g->g_cin));
				goto checkifs;	/* kludge */

			case READ_ENVDESCR:
				dprintf(("%s: \"ciscoEnvPresent\" is absent in this MIB\n", g->g_name));
				goto ifmem;	/* kludge */

			case READ_MEM:
				dprintf(("%s:got error in mem test%d\n", g->g_name, g->g_cin));
				for (m = g->g_mem; m; m = m->mem_next) {
				  if (m->mem_type == g->g_cin) {
				    m->mem_valid = -1;
			            dprintf(("MATCH BAD MEM POLL %s\n", mempolls[m->mem_type]));
				    break;
				  }
			        }
				goto checkmem;

			case POLL_IF:
                                dprintf(("%s: bad POLL_IF request\n",g->g_name));
				if (g->g_nwait)
					g->g_nwait--;	/* we must do it to
							 * avoid of lossing
							 * SNMP_ACTION */
				if (g->g_nwait == 0) {
				  g->g_act.a_time += g->g_poll ? g->g_poll : poll_period;
				  g->g_act.a_f = poll_send;
				  timeout((struct ACTIVE *) g);
				}
                                break;

			case READ_BGP:
				dprintf(("%s: bad bgp READ_BGP request\n", g->g_name));
				goto dopollif;

			default:	/*       :-) */
			}
		}
		fprintf(stderr, "%s: error in packet from %s request is %d: %s \n",
			av0, g->g_name, g->g_ps, snmp_errstring(pdu->errstat));
		return 1;
	}
	/* The first variable is always system.sysUptime */
	var = pdu->variables;

	/* GW has been reanimated */
	if (g->g_alive == G_DOWN) {
		j = g->g_alive;
		g->g_alive = G_UP;
		/* g->g_sess->retries = SNMP_RETRIES; */
		gw_opstat(g, 1);
		g->g_lastchst = time(0);
		dprintf(("%s : GW IS ALIVE\n", g->g_name));
	}
	/* Process the reply */
	switch (g->g_ps) {
	case READ_IFNUM:
		/*
		 * Retrieve the number of gateway's interfaces
		 */
		GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( check_uptime(lo,g) ) {
                  poll_send((struct ACTIVE *) g);
                  break;
                }
                g->g_uptime = lo;

		NEXTVAR;
		g->g_alive = G_UP;
		GETVAR(g->g_ngwifs, "ifNumber", ASN_INTEGER, break);
		NEXTVAR;
		if (g->g_ngwifs < 0) {
			fprintf(stderr, "%s: wrong  %d for ifNumber from %s\n",
				av0, var->type, g->g_ngwifs, g->g_name);
			break;
		}
		dprintf(("%s: %d interfaces\n", g->g_name, g->g_ngwifs));

		/*
		 * Read the product ID code
		 */
		GETVAR(j, "sysObjID", ASN_OBJECT_ID, break);
#ifdef DEBUG
		if (debug) {
			printf("%s SysObjID=", g->g_name);
			for (j = 0; j < var->val_len / sizeof(oid); j++)
				printf(".%d", var->val.objid[j]);
			printf("\n");
		}
#endif
		if (var->val_len == sizeof v_ciscoProductID &&
		    !memcmp(var->val.objid, v_ciscoProductID, sizeof(v_ciscoProductID) - sizeof(oid) )) {
			g->g_cisco = 1;
			g->g_ps = READ_ENVDESCR;
		} else
			g->g_ps = MAP_STATE[0];
		g->g_cin = 0;
		poll_send((struct ACTIVE *) g);
		break;

	case READ_ENVDESCR:
		/*
		 * We can use only revision 2 or 3 of the cisco environmental
		 * card
		 */
		GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( check_uptime(lo,g) ) {
                  poll_send((struct ACTIVE *) g);
		  break;
                }
                g->g_uptime = lo;

		NEXTVAR;
		GETVAR(j, "ciscoEnvPresent", ASN_INTEGER, goto ifmem);
		dprintf(("%s cisco EnvPresent=%d\n", g->g_name, j));
		if (j == 2 || j == 3)
			g->g_ciscoenv = 1;

ifmem:

		if(g->g_mem) { 
		       /* Now - reset all memory types */
		       for (m = g->g_mem; m; m = m->mem_next)
			       m->mem_valid = MEM_VALID_BAD;
		       g->g_cin = g->g_mem->mem_type;
		       g->g_ps = READ_MEM;
		}
	        else  
		  g->g_ps = MAP_STATE[0];

		poll_send((struct ACTIVE *) g);
		break;

	case READ_MEM:
                GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( check_uptime(lo,g) ) {
                  poll_send((struct ACTIVE *) g);
                  break;
                }
                g->g_uptime = lo;

                NEXTVAR;
		for (m = g->g_mem; m; m = m->mem_next) {
			if (m->mem_type == g->g_cin) {
		/* very strange bug at some ciscos. I ask one, she answed with another */
			  if ( m->mem_type == var->name[13] ) {
				if(var != 0 ) {
				        m->mem_valid = MEM_VALID_OK;
			        	dprintf(("MATCH MEM POLL %s\n", mempolls[m->mem_type]));
				}
				else 
				  m->mem_valid = -1;
				break;
			   }
                           else
				 m->mem_valid = -1;
			   break;
                        }
		}
checkmem:
		/* All required mem polls */
		for (m = g->g_mem; m; m = m->mem_next) {
			if (m->mem_valid ==  MEM_VALID_BAD) {
				g->g_cin = m->mem_type;
				break;
			}
		}

		if(m==0) {
		  for (m = g->g_mem; m; m = m->mem_next) {
			if (m->mem_valid == -1)
				fprintf(stderr, "%s: mem poll %s not found on %s\n",
				av0, mempolls[m->mem_type], g->g_name);
			}
		   g->g_ps = MAP_STATE[0];
		   g->g_cin = 0;
		}
	        else 
		g->g_ps = READ_MEM;

		poll_send((struct ACTIVE *) g);
		break;

	case READ_BGP:
		{
			char            buff[16];
			long            o;
			int             ok = 0;

			bzero(&buff, sizeof(buff));
			/* Try to fill BGP struct */
			if (var->type != IPADDRESS || var->name_length != 14)
				goto dopollif;
			sprintf(buff, "%d.%d.%d.%d",
				var->name[10], var->name[11], var->name[12], var->name[13]);

			for (b = g->g_bgp; b; b = b->bgp_next) {
				if (memcmp(b->bgp_ip, buff, strlen(buff)))
					continue;
				b->bgp_valid = BGP_VALID_OK;
				ok = 1;
				dprintf(("MATCH BGP PEER %s\n", b->bgp_ip));
				break;
			}

			/* add bgp peer */
			if (g->g_bgpall && !ok) {
				struct hostent *hp;
				struct in_addr  gia;
				struct BGP     *bn;
				char           *q;

				gia.s_addr = inet_addr(buff);
				if (gia.s_addr == -1)
					goto endbgp;

				for (b = g->g_bgp; b; b = b->bgp_next)
					if (b->bgp_next == NULL)
						break;

				bn = (struct BGP *) malloc(sizeof(struct BGP));
				if (!bn) {
					fprintf(stderr, "%s: cannot malloc for BGP %s\n",
						av0, g->g_name);
					exit(1);
				}
				bzero(bn, sizeof(struct BGP));
				bn->bgp_ip = (char *) strdup(&buff);
				/* try to get a name from DNS */
				hp = gethostbyaddr(&gia, sizeof(gia), AF_INET);
				if (hp) {
					if (q = index(hp->h_name, '.'))
						*q = 0;
					bn->bgp_name = (char *) malloc(strlen(hp->h_name) + strlen(g->g_sname) + 2);
					if (!bn->bgp_name) {
						fprintf(stderr, "%s: cannot malloc for BGP name %s\n",
							av0, g->g_name);
						exit(1);
					}
					strcpy(bn->bgp_name, g->g_sname);
					strcat(bn->bgp_name, "_");
					strcat(bn->bgp_name, hp->h_name);
				} else
					bn->bgp_name = bn->bgp_ip;
				bn->bgp_next = NULL;
				if (g->g_bgp)
					b->bgp_next = bn;
				else
					g->g_bgp = bn;
				bn->bgp_reqid = 0;
				bn->bgp_as = 0;
				bn->bgp_valid = BGP_VALID_OK;
				bn->bgp_state = BGP_IDLE;
				bn->bgp_time = 0;
				dprintf(("ADD NEW BGP PEER %s %s\n", bn->bgp_ip, bn->bgp_name));
			}
	endbgp:

			if (!g->g_bgpall) {
				for (b = g->g_bgp; b; b = b->bgp_next) {
					if (b->bgp_valid != BGP_VALID_OK)
						break;
				}
				/* All bgp peers are BGP_VALID_OK */
				if (b == 0)
					goto dopollif;
			}
			/* Poll next BGP */
			g->g_ps = READ_BGP;
			bcopy((char *) var->name, (char *) g->g_oidnext,
			      var->name_length * sizeof(oid));
			g->g_oidlen = var->name_length;
			poll_send((struct ACTIVE *) g);
		}
		break;
dopollif:
		/* Do POLL_IF request */
		g->g_ps = POLL_IF;	/* try to know interface type */
		j = poll_period / ngateways;
		if (j < 1)
			j = 1;
		g->g_act.a_time = tsharp(time(0),
					 g->g_poll ? g->g_poll : poll_period,
					 j * g->g_n);
		g->g_act.a_f = poll_send;
		dprintf(("router %s poll_period is %d\n", g->g_name, j));
		timeout((struct ACTIVE *) g);
		break;
	case POLL_UP:
		/* response after gateway was in INAC state */
		GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( !check_uptime(lo,g) ) {
			/* router must be up after INAC */
			dprintf(("%s: GW IS INAC/UP\n", g->g_name));
			if (g->g_alive == G_FIRST)
				reset_gwdata(g);
			else {
				g->g_ps = READ_IFNUM;
				reset_reqid(g);
			};

			g->g_uptime = lo;
			g->g_alive = G_UP;
			gw_opstat(g, 1);
			g->g_lastchst = time(0);
		}
		poll_send((struct ACTIVE *) g);
		break;

	case READ_IFINDEX:
		GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( check_uptime(lo,g) ) {
                  poll_send((struct ACTIVE *) g);
                  break;
                }
		g->g_uptime = lo;

		for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
			if (i->i_map.m_key == READ_IFINDEX) {
				i->i_index = atoi(i->i_map.m_val);
			}
		}
		goto END_OF_MAP;

	case READ_IFDESCR:
		/*
		 * Retrieve the list of interface names and match them
		 * against specifications in config file.
		 */

		if (var->name_length != 11) {
			dprintf(("%s: Bad var var->val.string %s, var->name_length %d\n",
				 av0, var->val.string, var->name_length));
			goto checkifs;
		}
		GETVAR(j, "ifDescr", ASN_OCTET_STR, goto checkifs);
		dprintf(("%s: if%d -> '%.*s'\n", g->g_name, g->g_cin, var->val_len, var->val.string));
		/* There is no interface to monitor */
		if (g->g_nifs == 0)
			goto checkifs;

		/*
		 * Check if the interface is in configuration list with that
		 * MAP
		 */

		for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
			int l = strlen(i->i_name);
			char *pi;
			/* If there is space in description, cut the tail */
			register int ii;
			for ( ii = 1; ii < var->val_len; ii++) {
			       if ( var->val.string[ii] == ' ' ) {
				       var->val.string[ii] = 0;
				       var->val_len = ii;
				       break;
			       };
			};
			/* Check if it's subinterface */
			pi = index(i->i_name,'.');
			if (l != var->val_len && (!pi ||  var->val_len < l+5) )
				continue;
			if (i->i_map.m_key != READ_IFDESCR)
				continue;
			if (memcmp(i->i_name, var->val.string, var->val_len))
			{
				/* Recognize ATM1/0.1-aal5 level - IOS 12.0 */
				/* 1. there is not '.' in the name */
				if ( !pi ) continue;
				/* 2. descr is not long enougph */
				if ( var->val_len < l + 5 ) continue;
				/* 3. The head part of name do not compare */
				if ( memcmp(i->i_name,var->val.string,strlen(i->i_name) ) ) continue;
				/* 4. The tail is not -aal5 */
				if ( memcmp(var->val.string + strlen(i->i_name),"-aal5",5) ) continue;
			};
			i->i_index = var->name[10];
			dprintf(("Found %s %d\n",i->i_name,i->i_index));
			break;
		}
		/* If there are more interfaces, poll them */
		g->g_cin = var->name[10];
		dprintf(("index %d\n", j));
		poll_send((struct ACTIVE *) g);
		break;

checkifs:
		/*
		 * Check that all interfaces from config file were found
		 */
		for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
			if (i->i_index < 0) {
				/*
				 * there are another MAP function to map
				 * IfIndex
				 */
				g->g_ps = MAP_STATE[1];
				poll_send((struct ACTIVE *) g);
				goto END;
			}
		}
END_OF_MAP:
		g->g_ps = READ_IFTYPE;
		poll_send((struct ACTIVE *) g);
		break;

	case READ_IFTYPE:
		/* Calculate interface type */

		GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( check_uptime(lo,g) ) {
                  poll_send((struct ACTIVE *) g);
                  break;
                }
                g->g_uptime = lo;

		NEXTVAR;
		for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
			if (i->i_index < 0)
				continue;
			GETVAR(j, "ifType", ASN_INTEGER, return 1);
			NEXTVAR;
			i->i_type = j;
			dprintf(("%s %s type %d\n", g->g_name, i->i_name, i->i_type));
		}

		for (i = &g->g_ifs[g->g_nifs - 1]; i >= g->g_ifs; i--) {
			if (index(i->i_name, '.') != NULL && i->i_type != ifType_aal5 )
			{
				dprintf(("force type to FR: %s %d\n",i->i_name,i->i_type));
				i->i_type = ifType_frame_relay;
			}
			if (i->i_index < 0) {
				fprintf(stderr, "%s: interface %s not found on %s\n",
					av0, i->i_name, g->g_name);
			}
		}

		/* Try to get BGP bgpPeerIdentifier */
		if (g->g_bgp || g->g_bgpall) {
			g->g_ps = READ_BGP;
			poll_send((struct ACTIVE *) g);
			break;
		}
		/*
		 * Now i don't know what should we do with interface which
		 * has only small spare interface statisctic. For instance,
		 * is Frame relay subinterface in Cisco. sertanly, we can
		 * determinate type of interface and poll only known MIBS for
		 * it, but what about mainFrInterface with has all MIB.
		 * Nowday there is no CISCO-FRAME-RELAY-MIB.my at cisco box.
		 * If we poll small packet of MIB and determinate what
		 * interface is it, i will be very lavish do one poll for one
		 * interface. So i try to deteminate subinterface. Read
		 * poll.c.fr
		 */

		/* Try to get BGP bgpPeerIdentifier if bgp is requested */
		if (g->g_bgp || g->g_bgpall) {
			g->g_ps = READ_BGP;
			poll_send((struct ACTIVE *) g);
			break;
		}
                else 
                    goto dopollif;
END:
		break;
	case POLL_IF:
		/*
		 * Gateway responce
		 */
		GETVAR(lo, "sysUptime", TIMETICKS, return 1);
                if ( check_uptime(lo,g) ) {
                   poll_send((struct ACTIVE *) g);
                   break;
                }
		g->g_uptime = lo;

		NEXTVAR;
		switch (iff) {
		case RECV_GAT:
			dprintf(("%s: GW RESP (nwait=%d)", g->g_name, g->g_nwait));
			if (g->g_cisco) {
				g->g_stat.gs_cnt++;
				GETVAR(j, "ciscoAvgBusy", ASN_INTEGER, goto dnw);
				dprintf((" cpu=%d", j));
				if (j > 100)
					j = 100;
				if (j < 0)
					j = 0;
				g->g_cpubusy = j;
				if (g->g_cpubusy > gw_maxcpu) {
					if (!g->g_cpubusytime)
						g->g_cpubusytime = time(0);
				} else
					g->g_cpubusytime = 0;
				g->g_stat.gs_cpuavg += j;
				if (g->g_stat.gs_cpumax < j)
					g->g_stat.gs_cpumax = j;
				NEXTVAR;
				GETVAR(j, "ciscoFreeMem", ASN_INTEGER, goto dnw);
				dprintf((" freemem=%d", j));
				if (j < 0)
					j = 0;
				j /= 1024;
				g->g_freemem = j;
				if (g->g_freemem < gw_minmem) {
					if (g->g_memlowtime == 0)
						g->g_memlowtime = time(0);
				} else
					g->g_memlowtime = 0;
				g->g_stat.gs_memavg += j;
				if (g->g_stat.gs_memmin > j)
					g->g_stat.gs_memmin = j;
				if (g->g_ciscoenv) {
					NEXTVAR;
					GETVAR(j, "ciscoEnvTemp", ASN_INTEGER, goto dnw);
					dprintf((" temp=+%d C", j));
					if (j > g->g_stat.gs_temp)
						g->g_stat.gs_temp = j;
				}
			}
			dprintf(("\n"));
			break;

		case RECV_BGP:
			/* Calculate BGP state for b */

			dprintf(("%s: BGP RESP %s %s (nwait=%d)\n",
				 g->g_name, b->bgp_ip, b->bgp_name, g->g_nwait));
			bzero((char *) &nb, sizeof(nb));
			GETVAR(nb.bgp_state, "bgpPeerState", ASN_INTEGER, goto dnw);
			NEXTVAR;
			GETVAR(nb.bgp_as, "bgpPeerRemoteAs", ASN_INTEGER, goto dnw);
			NEXTVAR;
			GETVAR(nb.bgp_time, "bgpPeerFsmEstablishedTime", GAUGE, goto dnw);

			dprintf(("bgpPeerState=%d, bgpPeerRemoteAs = %d, bgpPeerFsmEstablishedTime = %lu\n", nb.bgp_state, nb.bgp_as, nb.bgp_time));

			if (nb.bgp_state < BGP_IDLE ||
			    nb.bgp_state > BGP_ESTABLISHED) {
				fprintf(stderr, "%s: (%s) bad bgpPeerState %s value (%d)\n",
				 av0, g->g_name, b->bgp_name, nb.bgp_state);
				goto dnw;
			}
			/* Compute new bgp parameters */
			bgp_stat(g, b, &nb);
			break;
		case RECV_MEM:
			/* Calculate Mem poll state for m */
			m->mem_cnt++;
			GETVAR(j, "ciscoMemoryPoolUsed", GAUGE, goto dnw);
			NEXTVAR;
			if (j < 0)
				j = 0;
			j /= 1024;
			m->mem_used = j;
			if (j > m->mem_maxused) 
				m->mem_maxused = j;
			m->mem_avgused +=j;

			GETVAR(j, "ciscoMemoryPoolFree", GAUGE, goto dnw);
			NEXTVAR;
			if (j < 0)
				j = 0;
			j /= 1024;
			m->mem_free = j;
			if (m->mem_minfree < j)
				m->mem_minfree = j;
			m->mem_avgfree += j;

			GETVAR(j, "ciscoMemoryPoolLargestFree", GAUGE, goto dnw);
			if (j < 0)
				j = 0;
			j /= 1024;
			m->mem_maxseg = j;
			if (m->mem_minmaxseg < j)
                            m->mem_minmaxseg = j;
			m->mem_avgmaxseg += j;

			dprintf(("%s: MEM RESP %s Used = %lu, Free = %lu, LargestFree = %lu (nwait=%d)\n", g->g_name, mempolls[m->mem_type], m->mem_used, m->mem_free, m->mem_maxseg,g->g_nwait ));

			if (m->mem_used < m->mem_minmem) {
				if (m->mem_memlowtime == 0)
					m->mem_memlowtime = time(0);
			} else
				m->mem_memlowtime = 0;
			break;

		case RECV_INT:
			/*
			 * Interface responce
			 */
			bzero((char *) &l, sizeof(struct LSTAT));
			l.l_timestamp = g->g_uptime;
			l.l_reftime = time(0);
			dprintf(("%s:%s RESP (nwait=%d)\n", g->g_name, i->i_name, g->g_nwait));
			/* Parse the link status data */
			GETVAR(l.l_Speed, "ifSpeed", GAUGE, goto dnw);
			NEXTVAR;
			GETVAR(l.l_OperStatus, "ifOperStatus", ASN_INTEGER, goto dnw);
			NEXTVAR;
			GETVAR(l.l_LastChange, "ifLastChange", TIMETICKS, goto dnw);
			NEXTVAR;

			switch (i->i_type) {
			case ifType_frame_relay:
				GETVAR(l.l_InOctets, "ifInOctets", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutOctets, "ifOutOctets", COUNTER, goto dnw);
				NEXTVAR;
				break;
			case ifType_aal5:
				GETVAR(l.l_InOctets, "ifInOctets", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_InUcastPkts, "ifInUcastPkts", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutOctets, "ifOutOctets", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutUcastPkts, "ifOutUcastPkts", COUNTER, goto dnw);
				NEXTVAR;
				break;
			default:
				GETVAR(l.l_InOctets, "ifInOctets", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_InUcastPkts, "ifInUcastPkts", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_InNUcastPkts, "ifInNUcastPkts", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_InDrops, "ifInDiscards", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_InErrors, "ifInErrors", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutOctets, "ifOutOctets", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutUcastPkts, "ifOutUcastPkts", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutNUcastPkts, "ifOutNUcastPkts", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutDrops, "ifOutDiscards", COUNTER, goto dnw);
				NEXTVAR;
				GETVAR(l.l_OutErrors, "ifOutErrors", COUNTER, goto dnw);
				break;
			};

			if (l.l_OperStatus < IF_UP || l.l_OperStatus > IF_TESTING) {
				fprintf(stderr, "%s: (%s) bad ifOperStatus value (%d)\n",
				 av0, g->g_name, i->i_name, l.l_OperStatus);
				goto dnw;
			}
			/* Compute new link parameters */
			link_stat(g, i, &l);
			break;
		};

		/* If all the replies received, go on with polling */
dnw:            if (--(g->g_nwait) <= 0) {
			g->g_act.a_time += g->g_poll ? g->g_poll : poll_period;
			g->g_act.a_f = poll_send;
			timeout((struct ACTIVE *) g);
		}
	}
	return 1;
}


/*
 * The input loop. Never returns.
 */
void
input_loop()
{
	int             nf;
	fd_set          fds;
	struct timeval  tv;
	int             block;
	int             cnt;

	for (;;) {
		nf = 0;
		FD_ZERO(&fds);
		if ((tv.tv_sec = sched()) != -1) {
			tv.tv_usec = 0;
			block = 0;
		} else {
			timerclear(&tv);
			block = 1;
		}
		snmp_select_info(&nf, &fds, &tv, &block);
		cnt = select(nf, &fds, 0, 0, &tv);
		if (cnt > 0) {
			snmp_read(&fds);
			continue;
		} else if (cnt == 0) {
			dprintf(("! "));
			snmp_timeout();
		}
	}
	/* NOTREACHED */
}

/* Check if router has got DOWN state */
check_uptime(uptime,g)
	long uptime;
        struct GATEWAY *g;
{ 
  if (g->g_uptime == 0) return 0;
  dprintf(("%s: uptime is %ld g->g_uptime is %ld\n", g->g_name, uptime, g->g_uptime));

  if (uptime < g->g_uptime - 2000) {
        /* it must be router was down and up */
	dprintf(("%s: GW IS INAC/DOWN/UP\n", g->g_name));
	/* We have to reset all statistic */
	reset_gwdata(g);
	g->g_alive = G_DOWN;
	gw_opstat(g, 1);
	g->g_alive = G_UP;
	gw_opstat(g, 2);
	g->g_lastchst = time(0);
        return 1;
  }
  return 0;
}
