/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: snmpplot.c,v $
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.13  1998/03/20  19:02:40  alex
 * RELCOM
 *
 * Revision 1.12  1996/01/10  15:32:19  tim
 * Common code removed to another file (pdate.c)
 *
 * Revision 1.11  1995/12/11  14:56:22  tim
 * *** empty log message ***
 *
 * Revision 1.10  1995/09/07  09:29:53  tim
 * ???
 *
 * Revision 1.9  1994/02/02  18:45:35  avg
 * Added support for BSD-style regexp library
 *
 * Revision 1.9  1994/02/02  18:45:35  avg
 * Added support for BSD-style regexp library
 *
 * Revision 1.8  1993/12/16  20:34:47  avg
 * Changed usage msg.
 *
 * Revision 1.7  1993/12/16  20:30:32  avg
 * Added check for case of no args.
 *
 * Revision 1.6  1993/09/29  20:58:26  avg
 * Fixed couple of bugs in formats.
 *
 * Revision 1.5  1993/09/22  23:25:04  avg
 * Added drawing CPU and memory utilization graphs for routers.
 *
 * Revision 1.5  1993/09/22  23:25:04  avg
 * Added drawing CPU and memory utilization graphs for routers.
 *
 * Revision 1.4  1993/09/02  20:46:57  avg
 * Fixed a bug with -l (didn't close files); now no
 * patterns means all links are ok; cleaned up -l printout.
 *
 * Revision 1.3  1993/08/13  17:51:08  avg
 *  Links will no longer be reported twice even if they
 * match several patterns.
 * Fixed starting date reset.
 * (Multi-link reports only).
 *
 * Revision 1.2  1993/08/11  19:40:10  avg
 * Added copyright.
 * Fixed file switching.
 * Now pictures will be in the specified order.
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 */
#ifndef lint
static char rcsid[] = "$Id: snmpplot.c,v 1.1.1.1 1999/07/19 18:38:31 alex Exp $";
#endif

#include "manifest.h"

#define MAXMEM 16384    /* Max router memory (kilobytes) */

/*
 * Plot the SNMP statistics data
 * using PostScript file format
 */
#include <ctype.h>

#ifndef REGEXPINCL
#define INIT    register char *bp = instring;
#define GETC()  (*bp++)
#define PEEKC() (*bp)
#define UNGETC(c) bp--
#define RETURN(c) return (c);
#define ERROR(c) regerr(c)
#include <regexp.h>
#else
#include REGEXPINCL
#endif

char *av0;

char expbuf[1024];

static mon_days[] = {
	31, 28, 31, 30,
	31, 30, 31, 31,
	30, 31, 30, 31
};

static char *mon_names[] = {
	"January", "February", "March", "April",
	"May", "June", "July", "August",
	"September", "October", "November", "December"
};

char *numeral();

char *arg_pp, *arg_ap;

FILE *open_linklog();

main(ac, av)
	int ac;
	char **av;
{
	extern char *optarg;
	extern int optind;
	FILE *prolog;
	register char c;
	char *cf = CONFIG_FILE;
	char *pf = PS_PROLOG_FILE;

	char *date;
	int yr, mo, dy; /* starting year/month/day */
	int cyr, cmo, cdy, hr;
	int i, ndays, nmatch, multifile;

#ifndef REGEXPINCL
	char *res[128]; /* Regular expressions */
#else
	regexp *res[128];
#endif
	char *re, *index(), *p;
	int nres, rn;
	char buf[512];

	enum FILE_CHANGE period = NONE;
	enum FILE_CHANGE rfc();

	struct GATEWAY *g;
	struct IF *ii;
	FILE *f;

	struct LAGG la;
	struct GAGG ga;
	struct tm tm;
/*adbr
        float speed, maxpps, tscale;
*/
        double speed;
        float maxpps, tscale;
	u_long tt, start_tt, end_tt;
	int hdrok, maxgwn;
	int ppscale;
	int pagenum = 1;

	int lflag = 0;
	int Lflag = 0, Rflag = 0;

	av0 = av[0];
	for(;;) {
		switch( getopt(ac, av, "LRlc:p:t:n:") ) {
		case -1:
			break;
		default:
		usage:
			fprintf(stderr, "Usage: %s [-L] [-R] [-l] [-c config_file] [-n page_number]\n\t[-p prolog_file] [-t period] year.month[.day] [name...]\n", av0);
			exit(1);

		case 'L':
			Lflag++;
			continue;

		case 'R':
			Rflag++;
			continue;

		case 'c':
			cf = optarg;
			continue;

		case 'p':
			pf = optarg;
			continue;

		case 't':
			period = rdfc(optarg);
			if( period < DAILY || period > MONTHLY ) {
				fprintf(stderr, "%s: bad period, valid periods are: day, week, month\n", av0);
				exit(1);
			}
			continue;

		case 'n':
			pagenum = xatoi(optarg);
			if( pagenum < 1 )
				goto usage;
			continue;

		case 'l':
			lflag++;
			continue;
		}
		break;
	}
	if( optind >= ac )
		goto usage;
	if( Lflag && Rflag ) {
		fprintf(stderr, "%s: options -R and -L are incompatible\n", av0);
		exit(1);
	}

	/*
	 * Parse date
	 */
	date = av[optind++];
	if (parsedate (date, &yr, &mo, &dy, &period)) {
		fprintf(stdout, "%s: bad date format (should be like 93.2 or 2001.11.03)\n", av0);
		exit(1);
	}
	/*
	 * Compile regular expressions
	 */
#ifndef REGEXPINCL
	re = expbuf;
	nres = 0;
	for( i = optind ; i < ac ; i++ ) {
		if( nres >= sizeof res / sizeof res[0] ) {
			fprintf(stderr, "%s: too many regular expressions\n", av[0]);
			exit(1);
		}
		res[nres++] = re;
		for( p = av[i] ; *p ; p++ ) {
			if( isupper(*p) )
				*p = tolower(*p);
		}
		re = compile(av[i], re, &expbuf[sizeof expbuf - 1], 0);
	}
#else
	nres = 0;
	for( i = optind ; i < ac ; i++ ) {
		for( p = av[i] ; *p ; p++ ) {
			if( isupper(*p) )
				*p = tolower(*p);
		}
		if( (res[nres++] = regcomp(av[i])) == 0 )
			exit(1);
	}
#endif /*REGEXPINCL*/

	/*
	 * Read in the configuration file
	 */
	if( rdconfig(cf) )
		exit(1);
	if( period == NONE )
		period = file_change;

	/*
	 * Copy the prolog file
	 */
	if( !lflag ) {
		if( (prolog = fopen(pf, "r")) == 0 ) {
			fprintf(stderr, "%s: cannot open PostScipt prolog file %s\n", av0, pf);
			exit(1);
		}
		while( (c = getc(prolog)) != EOF )
			putchar(c);
		fclose(prolog);
	}

	/*
	 * Generate timestamp records
	 */
	switch( period ) {
	case MONTHLY:
		ndays = mondays (mo, yr);
		dy = 1;
		break;
	case WEEKLY:
		ndays = 7;
		break;
	case DAILY:
		ndays = 1;
		break;
	}
	if( !lflag ) {
		printf("\n%%--- THE AUTOMATICALLY GENERATED DATA ---\n");
		printf("/timeStamps {\n");
		if( ndays == 1 ) {
			for( i = 0 ; i < 24 ; i++ ) {
				hr = i % 12;
				if( hr == 0 )
					hr = 12;
				printf("\t%f (%d:00 %s) timestamp\n",
					i/24.0, hr, i < 12 ? "am" : "pm");
			}
		} else {
			cyr = yr;
			cdy = dy;
			cmo = mo;
			for( i = 0 ; i < ndays ; i++ ) {
				printf("\t%f (%d/%d/%d) timestamp\n",
					((double)i)/ndays, cmo, cdy, cyr % 100);
				if( ++cdy > mondays (cmo, cyr) ) {
					cdy = 1;
					if( ++cmo > 12 ) {
						cmo = 1;
						cyr++;
					}
				}
			}
		}
		printf("} def\n\n");
	}

	/*
	 * Get the max width of gw names
	 */
	if( lflag ) {
		maxgwn = 0;
		for( g = gwlist ; g ; g = g->g_next ) {
			i = strlen(g->g_name);
			if( i > maxgwn )
				maxgwn = i;
		}
	}

	/*
	 * Calculate the start time
	 */
	tt = time(0);
	tm = *localtime(&tt);
	tm.tm_year = yr - 1900;
	tm.tm_mday = dy;
	tm.tm_mon = mo - 1;
	tm.tm_hour = 0;
	tm.tm_min = 0;
	tm.tm_sec = 0;
	start_tt = timelocal(&tm);
	end_tt = start_tt + ndays * 24 * 60 * 60;

	/*
	 * Scan all the routers for the pattern
	 */
	nmatch = 0;
	if( Lflag )
		goto skip_routers;
	multifile = 0;
	cyr = yr; cmo = mo; cdy = dy;
	for( rn = (nres?0:-1) ; rn < nres ; rn++ ) {
		for( g = gwlist ; g ; g = g->g_next ) {
			if( nres != 0 ) {
				re = buf;
				for( p = g->g_name ; *p ; p++ )
					*re++ = isupper(*p) ? tolower(*p) : *p;
				*re = 0;
#ifndef REGEXPINCL
				if( !step(buf, res[rn]) )
					continue;
#else
				if( !regexec(res[rn], buf) )
					continue;
#endif
			}
			if( g->g_alive++ )
				continue;       /* already processed! */
			nmatch++;
			yr = cyr; mo = cmo; cdy = dy;

			/* Open the router log file */
			hdrok = 0;
			f = open_linklog(g->g_name, yr, mo, dy);
			if( f == 0 ) {
				if( lflag )
				       fprintf(stderr, "Router %s: NO DATA AVAILABLE\n", g->g_name);
				else
				       fprintf(stderr, "%s: (warning) -- no log file %s/%d.%02d.%02d\n",
						av0, g->g_name, yr, mo, dy);
				continue;
			}

			/* Read the records (Utilization pass) */
			maxpps = 0.0;
			tscale = 1.0 / (ndays * 24 * 60 * 60);
		cont3:
			while( (i = rdrlog(f, &tt, &ga, start_tt, end_tt)) > 0 ) {
				if( i != IF_STAT )
					continue;
				if( hdrok == 0 ) {
					hdrok = 1;

					/*
					 * If we're printing the link listings
					 * show the link.
					 */
					if( lflag ) {
						printf("Router %s\n", g->g_name);
						goto rend_reached;
					}
					if( pagenum )
						printf("%%%%Page: %s %d\n", g->g_sname, pagenum);
					printf("/Router (%s) def\n", g->g_name);
					printf("/Time (%s", mon_names[mo-1]);
					if( period != MONTHLY )
						printf(" %d%s", dy, numeral(dy));
					printf(", %d) def\n", yr);
					printf("/AggTime (");
					outtime(aggr_period);
					printf(") def\n");
					printf("/PollTime (");
					outtime(g->g_poll? g->g_poll : poll_period);
					printf(") def\n\nrouterPage\n");
				}
				printf("%f %.3f %.3f %.3f %.3f d\n",
					(tt - start_tt) * tscale,
					ga.gs_cpuavg/100.0, ga.gs_cpumax/100.0,
					1.0 - ((double) ga.gs_memavg / MAXMEM),
					1.0 - ((double) ga.gs_memmin / MAXMEM));
			}
			if( i == -2 )
				goto rend_reached;

			/* Try to open the next log file */
			switch( file_change ) {
			case DAILY:
				if( period == DAILY )
					break;
			nextday3:
				if( ++dy > mondays (dy, mo) ) {
					dy = 1;
					if( ++mo > 12 ) {
						mo = 1;
						yr++;
					}
				}

				/* Do we need to open a new file? */
				tm.tm_year = yr - 1900;
				tm.tm_mday = dy;
				tm.tm_mon = mo - 1;
				tm.tm_hour = 0;
				tm.tm_min = 0;
				tm.tm_sec = 0;
				if( timelocal(&tm) >= end_tt )
					break;

				/* Try to open the new log file */
				fclose(f);
				f = open_linklog(g->g_name, yr, mo, dy);
				if( f != 0 )
					goto cont3;
				break;

			case WEEKLY:
				dy += 6;
				goto nextday3;

			case MONTHLY:
				if( period == DAILY || period == MONTHLY )
					break;
				dy += 100;
				goto nextday3;
			}

		rend_reached:
			if( hdrok == 0 ) {
				if( f != 0 )
					fclose(f);
				if( lflag )
				       fprintf(stderr, "Router %s: NO DATA AVAILABLE\n", g->g_name);
				continue;
			}
			if( lflag ) {
				if( f != 0 )
					fclose(f);
				continue;
			}
			if( pagenum )
				printf("\n(Page %d) pageNumber\n", pagenum++);
			printf("\nshowpage\n\n");
			if ( pagenum )
				printf("%%%%EndPage: %s %d\n", g->g_sname, pagenum);

		}
	}

	/*
	 * Scan all the links for the pattern
	 */
skip_routers:
	if( Rflag )
		goto skip_links;
	multifile = 0;
	cyr = yr; cmo = mo; cdy = dy;
	for( rn = (nres?0:-1) ; rn < nres ; rn++ ) {
		for( g = gwlist ; g ; g = g->g_next ) {
			for( ii = g->g_ifs ; ii < &g->g_ifs[g->g_nifs] ; ii++ ) {
				if( nres == 0 )
					goto matched;

				/* Compare for full.router.name:interface  */
				strcpy(buf, g->g_name);
				strcat(buf, ":");
				strcat(buf, ii->i_name);
				if( !strcasecmp(av[rn+optind], buf) )
					goto matched;

				/* Compare for short-router-name:interface */
				if( re = index(buf, '.') ) {
					*re = ':';
					strcpy(re+1, ii->i_name);
					if( !strcasecmp(av[rn+optind], buf) )
						goto matched;
				}

				/* Try to mactch file and comment fields */
				re = buf;
				for( p = ii->i_comment ; *p ; p++ )
					*re++ = isupper(*p) ? tolower(*p) : *p;
				*re = 0;
#ifndef REGEXPINCL
				if( step(buf, res[rn]) )
					goto matched;
#else
				if( regexec(res[rn], buf) )
					goto matched;
#endif
				re = buf;
				for( p = ii->i_logdir ; *p ; p++ )
					*re++ = isupper(*p) ? tolower(*p) : *p;
				*re = 0;
#ifndef REGEXPINCL
				if( step(buf, res[rn]) )
					goto matched;
#else
				if( regexec(res[rn], buf) )
					goto matched;
#endif
				continue;

			matched:
				if( ii->i_index++ )
					continue;       /* already processed! */
				nmatch++;
				yr = cyr; mo = cmo; cdy = dy;

				/* Open the link log file */
				hdrok = 0;
				f = open_linklog(ii->i_logdir, yr, mo, dy);
				if( f == 0 ) {
					if( lflag )
					       fprintf(stderr, "%*s -> %s: NO DATA AVAILABLE\n", maxgwn, g->g_name, ii->i_comment);
					else
					       fprintf(stderr, "%s: (warning) -- no log file %s/%d.%02d.%02d\n",
							av0, ii->i_logdir, yr, mo, dy);
					continue;
				}

				/* Read the records (Utilization pass) */
				maxpps = 0.0;
				tscale = 1.0 / (ndays * 24 * 60 * 60);
			cont1:
				while( (i = rdlog(f, &tt, &la,&speed, start_tt, end_tt)) > 0 ) {
					if( i != IF_STAT )
						continue;
					if( hdrok == 0 ) {
						hdrok = 1;

						/*
						 * If we're printing the link listings
						 * show the link.
						 */
						if( lflag ) {
							printf("%*s -> %s\n", maxgwn, g->g_name, ii->i_comment);
							goto end_reached;
						}
						if( pagenum ) {
							char *pnend;
							if (pnend = index (ii->i_comment, ' '))
								*pnend = 0;
							printf("%%%%Page: \"%s\" %d\n", ii->i_comment, pagenum);
							if (pnend)
								*pnend = ' ';
						}
						printf("/To (%s) def\n", ii->i_comment);
						printf("/From (%s) def\n", g->g_name);
						printf("/Time (%s", mon_names[mo-1]);
						if( period != MONTHLY )
							printf(" %d%s", dy, numeral(dy));
						printf(", %d) def\n", yr);
						printf("/Bandwidth (%.1f) def\n\n", speed);
						printf("/AggTime (");
						outtime(aggr_period);
						printf(") def\n");
						printf("/PollTime (");
						outtime(g->g_poll? g->g_poll : poll_period);
						printf(") def\n\nutilPage\n");
					}
					printf("%f %.3f %.3f %.3f %.3f d\n",
						(tt - start_tt) * tscale,
						la.s_inutil / 100, la.s_inpeakutil / 100,
						la.s_oututil / 100, la.s_outpeakutil / 100);
					if( maxpps < la.s_inpeakpckts )
						maxpps = la.s_inpeakpckts;
					if( maxpps < la.s_outpeakpckts )
						maxpps = la.s_outpeakpckts;
				}
				if( i == -2 )
					goto end_reached;

				/* Try to open the next log file */
				switch( file_change ) {
				case DAILY:
					if( period == DAILY )
						break;
				nextday1:
					if( ++dy > mondays (dy, mo) ) {
						dy = 1;
						if( ++mo > 12 ) {
							mo = 1;
							yr++;
						}
					}

					/* Do we need to open a new file? */
					tm.tm_year = yr - 1900;
					tm.tm_mday = dy;
					tm.tm_mon = mo - 1;
					tm.tm_hour = 0;
					tm.tm_min = 0;
					tm.tm_sec = 0;
					if( timelocal(&tm) >= end_tt )
						break;

					/* Try to open the new log file */
					multifile++;
					fclose(f);
					f = open_linklog(ii->i_logdir, yr, mo, dy);
					if( f != 0 )
						goto cont1;
					break;

				case WEEKLY:
					dy += 6;
					goto nextday1;

				case MONTHLY:
					if( period == DAILY || period == MONTHLY )
						break;
					dy += 100;
					goto nextday1;
				}

			end_reached:
				if( hdrok == 0 ) {
					if( f != 0 )
						fclose(f);
					if( lflag )
					       printf("%*s -> %s: NO DATA AVAILABLE\n", maxgwn, g->g_name, ii->i_comment);
					continue;
				}
				if( lflag ) {
					if( f != 0 )
						fclose(f);
					continue;
				}
				if( pagenum )
					printf("\n(Page %d) pageNumber\n", pagenum++);
				printf("\nshowpage\n\n");
				if ( pagenum ) {
					char *pnend;
					if (pnend = index (ii->i_comment, ' '))
						*pnend = 0;
					printf("%%%%EndPage: \"%s\" %d\n", ii->i_comment, pagenum);
					printf("%%%%Page: \"%s\" %d\n", ii->i_comment, pagenum);
					if (pnend)
						*pnend = ' ';
				}
				/* Generate the packet-per seconds scale */
				ppscale = outppsscale((int) maxpps);
				printf("ppsPage\n");

				/* Read the records again (Packets per second pass) */
				yr = cyr; mo = cmo; cdy = dy;
				if( !multifile )
					fseek(f, 0l, 0);
				else {
					if( f != 0 )
						fclose(f);
					f = open_linklog(ii->i_logdir, yr, mo, dy);
					if( f == 0 ) {
						fprintf(stderr, "%s: log file %s/%d.%02d.%02d disappeared\n",
							av0, ii->i_logdir, yr, mo, dy);
						goto pktskip;
					}
				}
			cont2:
				while( (i = rdlog(f, &tt, &la, &speed, start_tt, end_tt)) > 0 ) {
					if( i != IF_STAT )
						continue;
					printf("%f %.3f %.3f %.3f %.3f d\n",
						(tt - start_tt) * tscale,
						la.s_inpckts / ppscale, la.s_inpeakpckts / ppscale,
						la.s_outpckts / ppscale, la.s_outpeakpckts / ppscale);
				}
				if( i == -2 )
					goto pktskip;

				/* Try to open the next log file */
				if( multifile ) {
					switch( file_change ) {
					case DAILY:
					nextday2:
						if( ++dy > mondays (dy, mo) ) {
							dy = 1;
							if( ++mo > 12 ) {
								mo = 1;
								yr++;
							}
						}

						/* Try to open the new log file */
						fclose(f);
						f = open_linklog(ii->i_logdir, yr, mo, dy);
						if( f != 0 )
							goto cont2;
						break;

					case WEEKLY:
						dy += 6;
						goto nextday2;

					case MONTHLY:
						dy += 100;
						goto nextday2;
				       }
				}
			pktskip:
				if( pagenum )
					printf("\n(Page %d) pageNumber\n", pagenum++);
				printf("\nshowpage\n\n");
				if ( pagenum ) {
					char *pnend;
					if (pnend = index (ii->i_comment, ' '))
						*pnend = 0;
					printf("%%%%EndPage: \"%s\" %d\n", ii->i_comment, pagenum);
					if (pnend)
						*pnend = ' ';
				}
				if( f != 0 )
					fclose(f);
			}
		}
	}
skip_links:
	if( !lflag )
	       printf("%%%%EOF\n");

	if( nmatch == 0 ) {
		fprintf(stderr, "%s: no matching link/router names\n", av0);
		exit(1);
	}
	exit(0);
}

#ifndef REGEXPINCL
/*
 * Report an error in reg expr
 */
regerr(errno)
	int errno;
{
	char *p;

	switch( errno ) {
	default:
		fprintf(stderr, "%s: error %d in regular expression\n", av0, errno);
		exit(1);

	case 11: p = "Range endpoint too large."; break;
	case 16: p = "Bad number."; break;
	case 25: p = "``\\ digit'' out of range."; break;
	case 36: p = "Illegal or missing delimiter."; break;
	case 41: p = "No remembered search string."; break;
	case 42: p = "\\( \\) imbalance."; break;
	case 43: p = "Too many \\(."; break;
	case 44: p = "More than 2 numbers given in \\{ \\}."; break;
	case 45: p = "} expected after \\."; break;
	case 46: p = "First number exceeds second in \\{ \\}."; break;
	case 49: p = "[] imbalance."; break;
	case 50: p = "Regular expression(s) too long.";
	}
	fprintf(stderr, "%s: error in regular expression: %s\n", av0, p);
	exit(1);
}
#endif /*REGEXPINCL*/

/*
 * Produce numeral suffix
 */
char *
numeral(n)
	int n;
{
	n %= 100;
	if( (n / 10) == 1 )
		return "th";
	n %= 10;
	if( n == 1 )
		return "st";
	if( n == 2 )
		return "nd";
	if( n == 3 )
		return "rd";
	return "th";
}

/*
 * Generate the pps scale
 */
int
outppsscale(npps)
	int npps;
{
	int scale = 5, s1;
	int i, j;
	char *m;

	while( npps > scale ) {         /* 5?? */
		scale *= 2;
		if( npps <= scale )     /* 1?? */
			break;
		s1 = scale;
		scale *= 2;
		if( npps <= scale )     /* 2?? */
			break;
		scale = s1 * 3;         /* 3?? */
		if( npps <= scale )
			break;
		scale = s1 * 5;
	}
	printf("/PkScale [(0)");
	for( i = 1 ; i <= 10 ; i++ ) {
		j = i * scale;
		m = "";
		if( j >= 10000 ) {
			j /= 1000;
			m = "k";
			if( j >= 10000 ) {
				j /= 1000;
				m = "m";
			}
		}
		printf(" (%d", j / 10);
		if( j % 10 )
			printf(".%d", j % 10);
		printf("%s)", m);
	}
	printf("] def\n");
	return scale;
}

/*
 * Output the time period
 */
outtime(s)
	int s;
{
	struct tp {
		int udiv;
		int umax;
		char *uname;
	} *t;
	static struct tp tpp[] = {
		60,     300,    "sec",
		60,     300,    "min",
		24,     96,     "hour",
		1000000, 0,     "day",
	};

	for( t = tpp ; ; t++ ) {
		if( s < t->udiv || (s < t->umax && (s % t->udiv) != 0) ) {
			 printf("%d %s%s", s, t->uname, s > 1 ? "s" : "");
			 break;
		}
		s /= t->udiv;
	}
}
