/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: snmpreport-h.c,v $
 * Revision 1.1.1.1  1999/07/19 18:38:31  alex
 * First revision in cvs
 *
 * Revision 1.1  1998/03/20  19:02:40  alex
 * Initial revision
 *
 * Revision 1.14  1996/01/10  15:32:19  tim
 * Common code removed to another file (pdate.c)
 *
 * Revision 1.13  1995/09/07  09:29:53  tim
 * average line load
 *
 * Revision 1.13  1995/09/07  09:29:53  tim
 * average line load
 *
 * Revision 1.12  1995/08/04  14:25:02  tim
 * Added some output (avg line load etc.)
 *
 * Revision 1.11  1994/02/02  18:45:35  avg
 * Added support for BSD-style regexp library
 *
 * Revision 1.10  1994/01/03  22:08:20  avg
 * Added median availability and total data volume estimate
 * claculations.
 *
 * Revision 1.9  1994/01/03  18:01:06  avg
 * Forgot debugging output.
 *
 * Revision 1.8  1994/01/03  01:50:18  avg
 * Added -m option -- calculate MTBF and MTBR.
 *
 * Revision 1.7  1993/09/22  23:25:53  avg
 * Changed rdlog on rdrlog in router stats.
 *
 * Revision 1.6  1993/09/02  20:48:07  avg
 * Added -a for reproting the availability figures.
 *
 * Revision 1.5  1993/08/24  18:59:59  avg
 * Added reporting on router failures and improved
 * handling of the date in header.
 *
 * Revision 1.4  1993/08/13  17:51:08  avg
 * Links will no longer be reported twice even if they
 * match several patterns.
 *
 * Revision 1.3  1993/08/13  00:57:18  avg
 * Some stylistic changes.
 *
 * Revision 1.2  1993/08/12  23:20:33  avg
 * Added key -h.
 *
 * Revision 1.1  1993/08/11  19:40:10  avg
 * Initial revision
 *
 *
 */
#ifndef lint
static char rcsid[] = "$Id: snmpreport-h.c,v 1.1.1.1 1999/07/19 18:38:31 alex Exp $";
#endif

#include "manifest.h"

/*
 * Report the troubles in the network
 */
#include <ctype.h>
#include <string.h>

#ifndef REGEXPINCL
#define INIT    register char *bp = instring;
#define GETC()  (*bp++)
#define PEEKC() (*bp)
#define UNGETC(c) bp--
#define RETURN(c) return (c);
#define ERROR(c) regerr(c)
#include <regexp.h>
#else
#include REGEXPINCL
#endif

char *av0;

char expbuf[1024];

static mon_days[] = {
	31, 28, 31, 30,
	31, 30, 31, 31,
	30, 31, 30, 31
};

static char *mon_names[] = {
	"", "", "", "",
	"", "", "", "",
	"", "", "", ""
};

char *numeral();

char *arg_pp, *arg_ap;

FILE *open_linklog();

char *tod(), *saytime();

#define NDT     10      /* Number of reported down time periods */

struct DT {
	u_long at;
	u_long duration;
};

extern u_long down_at;

main(ac, av)
	int ac;
	char **av;
{
	extern char *optarg;
	extern int optind;
	FILE *prolog;
	register char c;
	char *cf = CONFIG_FILE;
	char *end_s = (char *)0;
# define PT_END if ( end_s ) { printf("%s\n",end_s); end_s = (char *)0; }
	char *date;
	int yr, mo, dy; /* starting year/month/day */
	int cyr, cmo, cdy, hr;
	int i, ndays, nmatch, ntroubled;
	int nrtroubled, nrmatch;

#ifndef REGEXPINCL
	char *res[128]; /* Regular expressions */
#else
	regexp *res[128];
#endif
	char *re, *index(), *p;
	int nres, rn;
	char buf[512];

	enum FILE_CHANGE period = NONE;
	enum FILE_CHANGE rfc();

	struct GATEWAY *g;
	struct IF *ii;
	FILE *f;

	struct LAGG la;
	struct GAGG ga;
	struct tm tm;
	u_long tt, start_tt, end_tt;
	int hr_start, hr_end;
	int maxgwn;

	int lflag = 0;
	int aflag = 0;
	int rflag = 0;

	char *m_ln;
	/* For EXACT name search */
	char *link_name = (char*)0;
	char *int_name  = (char*)0;
	char *rtr_name  = (char*)0;
	struct GATEWAY *ok_g, *m_g;

/*adbr
	float avg_errs, avg_drops, speed;
*/
        double speed;
        float avg_errs, avg_drops;
	float avg_inutil, avg_oututil;
	int n_inoverhead, n_outoverhead;
	int n_stat, n_errs, n_recs, n_drops;
	int n_down;
	u_long total_down;
	int ndt;
	struct DT dt[NDT+1];
	double tot;

	int r_cnt, r_maxtemp, r_n_temp, r_avgcpu, r_n_cpu, r_avgmem, r_n_mem;

	int mtbf = 0;
	int ever_up;    /* link/router was ever up */
	long last_fail;  /* last failure ended at */
	long l_up_sum, l_down_sum;
	int l_mean_ctr;
	long r_up_sum, r_down_sum;
	int r_mean_ctr;
	double data_total;

	av0 = av[0];
	hr_start = 0;
	hr_end = 24;
	for(;;) {
		switch( getopt(ac, av, "arlmc:h:t:d:") ) {
		case -1:
			if( optind < ac )
				break;
		default:
		usage:
			fprintf(stderr, "Usage: %s [-a] [-r] [-l] [-m] [-d dir_name] [-c config_file] [-h hour-hour] [-t period] year.month[.day] [name...]\n", av0);
			exit(1);

		case 'c':
			cf = optarg;
			continue;

		case 'm':
			mtbf = 1;
			continue;

		case 't':
			period = rdfc(optarg);
			if( period < DAILY || period > MONTHLY ) {
				fprintf(stderr, "%s: bad period, valid periods are: day, week, month\n", av0);
				exit(1);
			}
			continue;

		case 'h':
			p = optarg;
			hr_start = iget(&p);
			if( *p++ != '-' || hr_start < 0 || hr_start > 23 )
				goto usage;
			hr_end = iget(&p);
			if( hr_end == 0 )
				hr_end = 24;
			if( *p != 0 || hr_end <= hr_start )
				goto usage;
			continue;

		case 'a':
			aflag++;
			continue;

		case 'r':
			rflag++;
			continue;

		case 'l':
			lflag++;
			continue;
		}
		break;
	}

	/*
	 * Parse date
	 */
	date = av[optind++];
	if (parsedate (date, &yr, &mo, &dy, &period)) {
		fprintf(stdout, "%s: bad date format (should be like 93.2 or 2001.11.03)\n", av0);
		exit(1);
	}
	/*
	 * Compile regular expressions
	 */
#ifndef REGEXPINCL
	re = expbuf;
	nres = 0;
	for( i = optind ; i < ac ; i++ ) {
		/* R=name - THIS router */
		if ( av[i][0] == 'R' && av[i][1] == '=' ) {
		       rtr_name = & (av[i][2]);
		       break;
		};
		/* L=name - THIS link */
		if ( av[i][0] == 'L' && av[i][1] == '=' ) {
		       link_name = & (av[i][2]);
		       break;
		};
		/* I=name:name - THIS interface */
		if ( av[i][0] == 'I' && av[i][1] == '=' ) {
		       int_name = & (av[i][2]);
		       break;
		};
		if( nres >= sizeof res / sizeof res[0] ) {
			fprintf(stderr, "%s: too many regular expressions\n", av[0]);
			exit(1);
		}
		res[nres++] = re;
		for( p = av[i] ; *p ; p++ ) {
			if( isupper(*p) )
				*p = tolower(*p);
		}
		re = compile(av[i], re, &expbuf[sizeof expbuf - 1], 0);
	}
#else
	nres = 0;
	for( i = optind ; i < ac ; i++ ) {
		/* R=name - THIS router */
		if ( av[i][0] == 'R' && av[i][1] == '=' ) {
		       rtr_name = & (av[i][2]);
		       break;
		};
		/* L=name - THIS link */
		if ( av[i][0] == 'L' && av[i][1] == '=' ) {
		       link_name = & (av[i][2]);
		       break;
		};
		/* I=name:name - THIS interface */
		if ( av[i][0] == 'I' && av[i][1] == '=' ) {
		       int_name = & (av[i][2]);
		       break;
		};
		for( p = av[i] ; *p ; p++ ) {
			if( isupper(*p) )
				*p = tolower(*p);
		}
		if( (res[nres++] = regcomp(av[i])) == 0 )
			exit(1);
	}
#endif /*REGEXPINCL*/


	/*
	 * Read in the configuration file
	 */
	if( rdconfig(cf) )
		exit(1);

	/*
	 * If -d DIR_NAME option exist, use this name instead
	 * No doubt, there is a few recirds in gwlist, we use the first one here
	 */
	/*
	 * Generate timestamp records
	 */
	switch( period ) {
	case MONTHLY:
		hr_start = 0;
		hr_end = 24;
		ndays = mondays (mo, yr);
		dy = 1;
		break;
	case WEEKLY:
		hr_start = 0;
		hr_end = 24;
		ndays = 7;
		break;
	case DAILY:
		ndays = 1;
		break;
	}

	/*
	 * Get the max width of gw names
	 */
	if( lflag ) {
		maxgwn = 0;
		for( g = gwlist ; g ; g = g->g_next ) {
			/*  g_sname(g_name) */
			if (strcmp (g->g_name, g->g_sname))
			    i = strlen(g->g_name) + strlen (g->g_sname) + 2;
			else
			    i = strlen(g->g_name);
			if( i > maxgwn )
				maxgwn = i;
		}
	}

	/*
	 * Calculate the start time
	 */
	tt = time(0);
	tm = *localtime(&tt);
	tm.tm_year = yr - 1900;
	tm.tm_mday = dy;
	tm.tm_mon = mo - 1;
	tm.tm_hour = 0;
	tm.tm_min = 0;
	tm.tm_sec = 0;
	start_tt = timelocal(&tm);
	end_tt = start_tt + ndays * 24 * 60 * 60;
	start_tt += hr_start * 60 * 60;
	end_tt -= (24 - hr_end) * 60 * 60;
	tot = end_tt - start_tt;

	/*
	 * Print the header
	 */
	if( !lflag ) {
		printf("<HTML><HEAD><TITLE>Network/Links Reliability Report</TITLE></HEAD><BODY>\n");
		printf("<H3>  ");
		switch( period ) {
		case WEEKLY:
			printf(",  ");
			printf("%02d.%02d.%d", dy, mo, yr);
			if( hr_start != 0 || hr_end != 24 ) {
				printf(",  %d  %d ", hr_start, hr_end);
			}
			break;
			/*FALL THRU*/

		case DAILY:
			printf(" %02d.%02d.%d", dy, mo, yr);
			if( hr_start != 0 || hr_end != 24 ) {
				printf(",  %d  %d ", hr_start, hr_end);
			}

		case MONTHLY:
			printf(" %s, %d", mon_names[mo-1], yr);
		}
		if( end_tt > tt ) {
			tm = *localtime(&tt);
			printf(", ");
			switch( period ) {
			case WEEKLY:
				if( mo-1 != tm.tm_mon )
					printf(" %s", mon_names[tm.tm_mon]);
				/*FALL THRU*/

			case MONTHLY:
				printf(" %02d", tm.tm_mday);
				if( yr - 1900 != tm.tm_year )
					printf(", %d", tm.tm_year + 1900);
				/*FALL THRU*/

			case DAILY:
				printf(" %d:%02d",tm.tm_hour,tm.tm_min);
			}
		}
		printf("</H3>\n");
	}
	if( end_tt > tt )
		end_tt = tt;

	/*
	 * Initialize MTBF & MTBR counters
	 */
	l_up_sum = l_down_sum = 0;
	r_up_sum = r_down_sum = 0;
	r_mean_ctr = l_mean_ctr = 0;
	data_total = 0.0;

	/*
	 * Scan all the routers matching the pattern
	 */
	nrmatch = 0;
	nrtroubled = 0;
       if ( ! link_name &&  ! int_name )
	for( rn = (nres?0:-1) ; rn < nres ; rn++ ) {
		/* we'd like to get LINK or INTERFACE, not router */
		for( g = gwlist ; g ; g = g->g_next ) {
			/* If rtr_name is defined, don't check anything else */
			if ( rtr_name ) {
			       if ( ! strcasecmp(rtr_name,g->g_sname) )
				       goto ok;
			       continue;
			};
			if( nres ) {
				strcpy(buf, g->g_name);
				for( p = buf ; *p ; p++ ) {
					if( isupper(*p) )
						*p = tolower(*p);
				}
#ifndef REGEXPINCL
				if( !step(buf, res[rn]) )
					goto chk1;
#else
				if( !regexec(res[rn], buf) )
					goto chk1;
#endif
				goto ok;
chk1:
				strcpy(buf, g->g_sname);
				for( p = buf ; *p ; p++ ) {
					if( isupper(*p) )
						*p = tolower(*p);
				}
#ifndef REGEXPINCL
				if( !step(buf, res[rn]) )
					continue;
#else
				if( !regexec(res[rn], buf) )
					continue;
#endif
ok:
			}

			/* Open the router log file */
			cyr = yr; cmo = mo; cdy = dy;
			f = open_linklog(g->g_name, yr, mo, dy);
			if( f == 0 ) {
				if( lflag ) {
					if (strcmp(g->g_name, g->g_sname))
						fprintf(stderr, "%s(%s): NO DATA AVAILABLE\n", g->g_sname, g->g_name);
					else
						fprintf(stderr, "%s: NO DATA AVAILABLE\n", g->g_name);
				} else {
					if (strcmp(g->g_name, g->g_sname))
						fprintf(stderr, "%s(%s): (warning) -- no log file %s/%d.%02d.%02d\n",
							av0, g->g_sname, g->g_name, yr, mo, dy);
					else
						fprintf(stderr, "%s: (warning) -- no log file %s/%d.%02d.%02d\n",
							av0, g->g_name, yr, mo, dy);
				}
				continue;
			}

			/* Read the records */
			nrmatch++;
			down_at = (u_long)-1;
			n_down = 0;
			n_recs = 0;
			total_down = 0;
			r_cnt = 0;
			r_maxtemp = 0;
			r_n_temp = 0;
			r_avgcpu = 0;
			r_n_cpu = 0;
			r_avgmem = 0;
			r_n_mem = 0;
			bzero(dt, sizeof dt);
			last_fail = (u_long) -1;
			ever_up = 0;
		rcont:  while( (i = rdrlog(f, &tt, &ga, start_tt, end_tt)) > 0 ) {
				/*
				 * If we're printing the link listings
				 * show the link.
				 */
				n_recs++;
				if( lflag ) {
					if (strcmp(g->g_name, g->g_sname))
					    printf("<H4> <B>%s</B>(%s)</H4>\n", g->g_sname, g->g_name);
					else
					    printf("<H4> <B>%s</B></H4>\n", g->g_name);
					goto rend_reached;
				}

				/* Process records */
				switch( i ) {
				case IF_UP:
					ever_up = 1;
					if( down_at == (u_long)-1 )
					       break;
					i = tt - down_at;
					if ( i == 0 ) i = 1;
					if( i >= 0 ) {
						n_down++;
						total_down += i;
						new_dt(dt, down_at, i);
						if( i > 120 && ever_up ) {
							r_mean_ctr++;
							r_down_sum += i;
							if( last_fail == (u_long)-1 )
								r_up_sum += down_at - start_tt;
						if ( i == 0 ) i = 10;
							else
								r_up_sum += down_at - last_fail;
							last_fail = tt;
						}
					}
					down_at = (u_long)-1;
					break;

				case IF_DOWN:
					if( down_at == (u_long)-1 )
						down_at = tt;
					break;

				case IF_STAT:
					r_cnt++;
					if( ga.gs_temp > r_maxtemp )
						r_maxtemp = ga.gs_temp;
					if( ga.gs_temp > gw_maxtemp )
						r_n_temp++;
					r_avgcpu += ga.gs_cpuavg;
					if( ga.gs_cpuavg > gw_maxcpu )
						r_n_cpu++;
					r_avgmem += ga.gs_memavg;
					if( ga.gs_memavg < gw_minmem )
						r_n_mem++;
					break;
				}
			}
			if( i == -2 )
				goto rend_reached;

			/* Try to open the next log file */
			switch( file_change ) {
			case DAILY:
				if( period == DAILY )
					break;
			rnextday:
				if( ++cdy > mondays (cmo, cyr) ) {
					cdy = 1;
					if( ++cmo > 12 ) {
						cmo = 1;
						cyr++;
					}
				}

				/* Do we need to open a new file? */
				tm.tm_year = cyr - 1900;
				tm.tm_mday = cdy;
				tm.tm_mon = cmo - 1;
				tm.tm_hour = 0;
				tm.tm_min = 0;
				if( timelocal(&tm) >= end_tt )
					break;

				/* Try to open the new log file */
				fclose(f);
				f = open_linklog(g->g_name, cyr, cmo, cdy);
				if( f != 0 )
					goto rcont;
				break;

			case WEEKLY:
				cdy += 6;
				goto rnextday;

			case MONTHLY:
				if( period == DAILY || period == MONTHLY )
					break;
				cdy += 100;
				goto rnextday;
			}

		rend_reached:
			if ( n_stat == 0 ) n_stat = 1;
			if( down_at != (u_long)-1 ) {
				n_down++;
				total_down += end_tt - down_at;
				new_dt(dt, down_at, end_tt - down_at);
			} else {
				if( last_fail == (u_long)-1 )
					r_up_sum += end_tt - start_tt;
				else
					r_up_sum += end_tt - last_fail;
			}

			if( n_down != 0 || aflag || (rflag && r_cnt) ||
			    r_n_mem || r_n_cpu || r_n_temp ) {
				nrtroubled++;
				PT_END;
				if (strcmp(g->g_name, g->g_sname))
				    printf("<H4> <B>%s(%s)</B></H4>\n<UL>\n", g->g_sname, g->g_name);
				else
				    printf("<H4> <B>%s</B></H4>\n<UL>\n", g->g_name);
				end_s = "</UL>";
			} else
				ok_g = g;
			if( n_down > 1 ) {
				printf("<LI>  : %d \n", n_down);
				printf("<OL>\n");
				for( i = 0 ; dt[i].duration > 0 ; i++ )
					printf("<LI>%s\t%s\n", tod(dt[i].at),
						saytime(dt[i].duration));
				if( n_down > NDT )
					printf("...");
				printf("</OL><P>  : %s.\n", saytime(total_down));
			} else if( n_down == 1 ) {
				printf("<LI>    %s   %s.\n",
					saytime(dt[0].duration), tod(dt[0].at));
			}

			if( !r_cnt )
				goto r_nostat;

			if( r_n_cpu ) {
				printf("<LI>    (%d%%)   %s ",
				       gw_maxcpu, saytime(aggr_period * r_n_cpu));
				printf("(%s ).\n", saytime(aggr_period));
			}
			if( rflag || r_n_cpu )
				printf("<LI>  : %d%%.\n", r_avgcpu / r_cnt);
			if( r_n_mem ) {
				printf("<LI>     (%.2fMb)   %s ",
				       gw_minmem / 1024.0, saytime(aggr_period * r_n_mem));
				printf("(%s ).\n", saytime(aggr_period));
			}
			if( rflag || r_n_mem )
				printf("<LI>  : %.2fMb.\n",
					(r_avgmem / 1024.0) / r_cnt);
			if( r_maxtemp && (rflag || r_n_temp) )
				printf("<LI>   : %dC.\n", r_maxtemp);
			if( r_n_temp ) {
				printf("<LI>   (%dC)   %s ",
				       gw_maxtemp, saytime(aggr_period * r_n_temp));
				printf("(%s ).\n", saytime(aggr_period));
			}
		r_nostat:
			if( aflag )
				printf("<LI>  : %.5f%%.\n", 100.0* (tot - total_down) / tot);
			if( n_recs == 0 ) {
				nrmatch--;
				if( lflag )
				       fprintf(stderr, "%s: NO DATA AVAILABLE\n", g->g_name);
			}
			PT_END;
			if( f != 0 )
				fclose(f);
		}
	}

	/*
	 * Scan all the links matching the pattern
	 */
	nmatch = 0;
	ntroubled = 0;
	if ( ! rtr_name )
	    for( rn = (nres?0:-1) ; rn < nres ; rn++ ) {
		for( g = gwlist ; g ; g = g->g_next ) {
			for( ii = g->g_ifs ; ii < &g->g_ifs[g->g_nifs] ; ii++ ) {
				if ( int_name ) {
				    /* Compare for router.name:interface  */
				    strcpy(buf, g->g_sname);
				    strcat(buf, ":");
				    strcat(buf, ii->i_name);
				    if( !strcasecmp(int_name, buf) )
					    goto matched;

				    /* Compare for router-addr:interface */
				    strcpy(buf, g->g_sname);
				    strcat(buf, ":");
				    strcat(buf, ii->i_name);
				    if( !strcasecmp(int_name, buf) )
					goto matched;
				    continue;
				};
				if ( link_name ) {
				       char *p;
				       strcpy(buf,ii->i_comment);
				       if ( p = strchr(buf,' ') )  *p  = 0;
				       if ( !strcasecmp(link_name,buf) )
					       goto matched;
				       continue;
				};
				if( nres == 0 )
					goto matched;

				/* Compare for full.router.name:interface  */
				strcpy(buf, g->g_name);
				strcat(buf, ":");
				strcat(buf, ii->i_name);
				if( !strcasecmp(av[rn+optind], buf) )
					goto matched;

				/* Compare for short-router-name:interface */
				strcpy(buf, g->g_sname);
				strcat(buf, ":");
				strcat(buf, ii->i_name);
				if( !strcasecmp(av[rn+optind], buf) )
					goto matched;
				/* Try to mactch file and comment fields */
				re = buf;
				for( p = ii->i_comment ; *p ; p++ )
					*re++ = isupper(*p) ? tolower(*p) : *p;
				*re = 0;
#ifndef REGEXPINCL
				if( step(buf, res[rn]) )
					goto matched;
#else
				if( regexec(res[rn], buf) )
					goto matched;
#endif
				re = buf;
				for( p = ii->i_logdir ; *p ; p++ )
					*re++ = isupper(*p) ? tolower(*p) : *p;
				*re = 0;
#ifndef REGEXPINCL
				if( step(buf, res[rn]) )
					goto matched;
#else
				if( regexec(res[rn], buf) )
					goto matched;
#endif
				continue;

			matched:
				if( ii->i_index++ )
					continue;       /* already processed! */

				/* Open the link log file */
				cyr = yr; cmo = mo; cdy = dy;
				f = open_linklog(ii->i_logdir, yr, mo, dy);
				if( f == 0 ) {
					if( lflag ) {
					       if (strcmp(g->g_name, g->g_sname))
						    fprintf(stderr, "%s(%s) -> %s: NO DATA AVAILABLE\n", g->g_sname, g->g_name, ii->i_comment);
					       else
						    fprintf(stderr, "%s -> %s: NO DATA AVAILABLE\n", g->g_name, ii->i_comment);
					} else {
					       fprintf(stderr, "%s: (warning) -- no log file %s/%d.%02d.%02d\n",
							av0, ii->i_logdir, yr, mo, dy);
					}
					continue;
				}

				/* Read the records */
				nmatch++;
				down_at = (u_long)-1;
				n_down = 0;
				total_down = 0;
				avg_errs = 0.0;
				avg_drops = 0.0;
				avg_inutil = 0.0;
				avg_oututil = 0.0;
				n_stat = 0;
				n_errs = 0;
				n_drops = 0;
				n_inoverhead = 0;
				n_outoverhead = 0;
				n_recs = 0;
				bzero(dt, sizeof dt);
				last_fail = (u_long) -1;
				ever_up = 0;
			cont1:  while( (i = rdlog(f, &tt, &la, &speed, start_tt, end_tt)) > 0 ) {
					n_recs++;
					/*
					 * If we're printing the link listings
					 * show the link.
					 */
					if( lflag ) {
						if (strcmp(g->g_name, g->g_sname)) {
						    int ln, ls;
						    ln = strlen(g->g_name);
						    ls = maxgwn - ls - 2;
						    printf("%*s(%*s) -> %s\n", ls, g->g_sname, ln, g->g_name, ii->i_comment);
						} else
						    printf("%*s -> %s\n", maxgwn, g->g_name, ii->i_comment);
						goto end_reached;
					}

					/* Process records */
					switch( i ) {
					case IF_UP:
						ever_up = 1;
						if( down_at == (u_long)-1 )
							continue;
						i = tt - down_at;
						if( i > 0 ) {
							n_down++;
							total_down += i;
							new_dt(dt, down_at, i);
							if( i > 120 && ever_up ) {
								l_mean_ctr++;
								l_down_sum += i;
								if( last_fail == (u_long)-1 )
									l_up_sum += down_at - start_tt;
								else
									l_up_sum += down_at - last_fail;
								last_fail = tt;
							}
						}
						down_at = (u_long)-1;
						break;

					case IF_DOWN:
					case IF_TESTING:
						if( down_at != (u_long)-1 )
							continue;
						down_at = tt;
						break;

					case IF_STAT:
						n_stat++;
						avg_errs += la.s_inerrors;
						avg_drops += (la.s_outdrops + la.s_indrops);
						if( (la.s_outdrops+la.s_indrops)  > drop_threshold )
							n_drops++;
						if( la.s_inerrors > err_threshold )
							n_errs++;
						avg_inutil += la.s_inutil;
						if (la.s_inutil > ii->i_maxinutil)
							n_inoverhead++;
						avg_oututil += la.s_oututil;
						if (la.s_oututil > ii->i_maxoututil)
							n_outoverhead++;
						data_total += (poll_period * speed / 8.0) * la.s_oututil;
					}
				}
				if( i == -2 )
					goto end_reached;

				/* Try to open the next log file */
				switch( file_change ) {
				case DAILY:
					if( period == DAILY )
						break;
				nextday1:
					if( ++cdy > mondays (cmo, cyr) ) {
						cdy = 1;
						if( ++cmo > 12 ) {
							cmo = 1;
							cyr++;
						}
					}

					/* Do we need to open a new file? */
					tm.tm_year = cyr - 1900;
					tm.tm_mday = cdy;
					tm.tm_mon = cmo - 1;
					tm.tm_hour = 0;
					tm.tm_min = 0;
					tm.tm_sec = 0;
					if( timelocal(&tm) >= end_tt )
						break;

					/* Try to open the new log file */
					fclose(f);
					f = open_linklog(ii->i_logdir, cyr, cmo, cdy);
					if( f != 0 )
						goto cont1;
					break;

				case WEEKLY:
					cdy += 6;
					goto nextday1;

				case MONTHLY:
					if( period == DAILY || period == MONTHLY )
						break;
					cdy += 100;
					goto nextday1;
				}

			end_reached:
				if ( n_stat == 0 ) n_stat = 1;
				if( down_at != (u_long)-1 ) {
					n_down++;
					total_down += end_tt - down_at;
					new_dt(dt, down_at, end_tt - down_at);
				} else {
					if( last_fail == (u_long)-1 )
						l_up_sum += end_tt - start_tt;
					else
						l_up_sum += end_tt - last_fail;
				}

				if( n_drops > 0 || n_errs > 0 || n_down != 0 || aflag ) {
					ntroubled++;
					PT_END;
					end_s = "</UL>";
					if (strcmp(g->g_name, g->g_sname))
					    printf("<H4> <B>%s</B>  <B>%s(%s)</B></H4>\n<UL>\n",
						ii->i_comment[0] ? ii->i_comment : ii->i_name, g->g_sname, g->g_name);
					else
					    printf("<H4> <B>%s</B>  <B>%s</B></H4>\n<UL>\n",
						ii->i_comment[0] ? ii->i_comment : ii->i_name, g->g_name);
				} else {
					m_ln = ii->i_comment[0] ? ii->i_comment : ii->i_name;
					m_g  = g;
				}
				if( n_down > 1 ) {
					printf("<LI>  : %d \n", n_down);
					printf("<OL>\n");
					for( i = 0 ; dt[i].duration > 0 ; i++ )
						printf("<LI>%s\t%s\n", tod(dt[i].at),
							saytime(dt[i].duration));
					if( n_down > NDT )
						printf("<LI>...\n");
					printf("</OL>\n<LI>   : %s.\n", saytime(total_down));
				} else if( n_down == 1 ) {
					printf("<LI>   %s   %s.\n",
						saytime(dt[0].duration), tod(dt[0].at));
				}
				if( n_errs > 0 ) {
					printf("<LI>    (%.2f%%)  \n", err_threshold);
					printf("%s ", saytime(aggr_period * n_errs));
					printf("(%s ),      : \n",
						      saytime(aggr_period));
					printf("%.2f%%.\n", avg_errs / n_stat);
				}
				if( n_drops > 0 ) {
					printf("<LI>    (%.2f%%)  \n", drop_threshold);
					printf("%s ", saytime(aggr_period * n_drops));
					printf("(%s ),    : \n",
						      saytime(aggr_period));
					printf("%.2f%%.\n", avg_drops / n_stat);
				}
				if (( ii->i_maxinutil > 0.0 || ii->i_maxoututil > 0.0 && tot > total_down)) {
					printf("<LI>  : %.2f%% ()  %.2f%% ().\n",
						avg_inutil/n_stat, avg_oututil/n_stat);
					if (n_inoverhead > ndays || n_outoverhead > ndays) {
						printf("<LI>    (%8.0f%% in/%8.0f%% out)  \n",
							ii->i_maxinutil, ii->i_maxoututil);
						printf("%s ", saytime(aggr_period * (n_inoverhead+n_outoverhead)) );
						printf("(%s ).\n", saytime(aggr_period));
					}
				}
				if( aflag )
					printf("<LI> : %.2f%%.\n", 100.0* (tot - total_down) / tot);
				if( n_recs == 0 ) {
					nmatch--;
					if( lflag )
					       fprintf(stderr, "%s -> %s: NO DATA AVAILABLE\n", g->g_name, ii->i_comment);
				}
				if( f != 0 )
					fclose(f);
			}
		}
	}

	if( nmatch == 0 && nrmatch == 0 ) {
		fprintf(stderr, "%s: no matching link/router names\n", av0);
		exit(1);
	}
	if( !lflag && !aflag ) {
		printf("\n");
		if( nrtroubled < nrmatch ) {
			if( nrmatch == nrtroubled+1 ) {
			    if (strcmp(ok_g->g_name, ok_g->g_sname))
				printf("<H4> <B>%s(%s)</B> ", ok_g->g_sname, ok_g->g_name);
			    else
				printf("<H4> %s</B> ", ok_g->g_name);
			} else
				printf("<H4> %s ",
				       nrtroubled ? " " : "");
			printf("       </H4>\n");
		}
		if( ntroubled == nmatch - 1 ) {
		    if (strcmp(m_g->g_name, m_g->g_sname))
			printf("<H4> <B>%s</B>  <B>%s(%s)</B>   .</H4>\n", m_ln, m_g->g_sname, m_g->g_name);
		    else
			printf("<h4> <B>%s</B>  <B>%s</B>   .\n", m_ln, m_g->g_name);
		} else if( ntroubled == 0 && nmatch != ntroubled )
			printf("<H4>     .</H4>\n");
		else
			if ( nmatch != ntroubled ) printf("<h4>      .</H4>\n");
	}
	if( !lflag && mtbf ) {
		printf("<H3> </H3><UL>\n");
		if( nrmatch )  {
			if( r_mean_ctr ) {
				printf("<LI>  MTBF = %s,\n", saytime(r_up_sum/r_mean_ctr));
				printf("MTTR = %s.\n", saytime(r_down_sum/r_mean_ctr));
			} else
				printf("<LI> ,     .\n");
			printf("<LI>    %.5f%%.\n",
				((double)r_up_sum / (double)(r_up_sum + r_down_sum)) * 100.0);
		}
		if( nmatch )  {
			if( l_mean_ctr ) {
				printf("<LI>  - MTBF = %s,\n", saytime(l_up_sum/l_mean_ctr));
				printf("MTTR = %s.\n", saytime(l_down_sum/l_mean_ctr));
			} else
				printf("<LI> ,    .\n");
			printf("Median link availability is %.5f%%.\n",
				((double)l_up_sum / (double)(l_up_sum + l_down_sum)) * 100.0);
		}
		printf("<P>\n");
		printf("<H4>  ,     - \n");
		if( data_total > 562949953421312.0 )
			printf("%.2f exabytes.\n", data_total/1125899906842624.0);
		else if( data_total > 549755813888.0 )
			printf("%.2f petabytes.\n", data_total/1099511627776.0);
		else if( data_total > 536870912.0 )
			printf("%.2f terabytes.\n", data_total/1073741824.0);
		else if( data_total > 524288.0 )
			printf("%.2f gigabytes.\n", data_total/1048576.0);
		else if( data_total > 512.0 )
			printf("%.2f megabytes.\n", data_total/1024.0);
		else
			printf("%.2f kilobytes.\n", data_total);
		printf(".lp\n");
		printf("<P>   %d   %d .\n",
			nrmatch, nmatch);
	}
	printf("</BODY></HTML>\n");
	exit(0);
}

#ifndef REGEXPINCL
/*
 * Report an error in reg expr
 */
regerr(errno)
	int errno;
{
	char *p;

	switch( errno ) {
	default:
		fprintf(stderr, "%s: error %d in regular expression\n", av0, errno);
		exit(1);

	case 11: p = "Range endpoint too large."; break;
	case 16: p = "Bad number."; break;
	case 25: p = "``\\ digit'' out of range."; break;
	case 36: p = "Illegal or missing delimiter."; break;
	case 41: p = "No remembered search string."; break;
	case 42: p = "\\( \\) imbalance."; break;
	case 43: p = "Too many \\(."; break;
	case 44: p = "More than 2 numbers given in \\{ \\}."; break;
	case 45: p = "} expected after \\."; break;
	case 46: p = "First number exceeds second in \\{ \\}."; break;
	case 49: p = "[] imbalance."; break;
	case 50: p = "Regular expression(s) too long.";
	}
	fprintf(stderr, "%s: error in regular expression: %s\n", av0, p);
	exit(1);
}
#endif /*REGEXPINCL*/

/*
 * Produce numeral suffix
 */
char *
numeral(n)
	int n;
{
	n %= 100;
	if( (n / 10) == 1 )
		return "";
	n %= 10;
	if( n == 1 )
		return "";
	if( n == 2 )
		return "";
	if( n == 3 )
		return "";
	return "";
}

/*
 * Output the time period
 */
outtime(s)
	int s;
{
	struct tp {
		int udiv;
		int umax;
		char *uname;
	} *t;
	static struct tp tpp[] = {
		60,     300,    "",
		60,     300,    "",
		24,     96,     "",
		1000000, 0,     "",
	};

	for( t = tpp ; ; t++ ) {
		if( s < t->udiv || (s < t->umax && (s % t->udiv) != 0) ) {
			 printf("%d %s%s", s, t->uname, s > 1 ? "" : "");
			 break;
		}
		s /= t->udiv;
	}
}

/*
 * Print the human-readable period of time.
 */
char *
saytime(t)
	u_long t;
{
	static char buf[128];
	char *bp = buf;
	int n;

#define M      60
#define H      (60*M)
#define D      (24*H)

	if( t >= D ) {
		n = t / D;
		t %= D;
		sprintf(bp, "%d %s", n, n != 1 ? "" : "");
		bp += strlen(buf);
		if( t > 2*H )
			t -= t % H;     /* get rid of minutes & seconds */
		else
			t -= t % M;     /* get rid of seconds */
	}
	if( t >= H ) {
		if( bp != buf )
			*bp++ = ' ';
		n = t / H;
		t %= H;
		sprintf(bp, "%d %s", n, (n%10 == 1 && n != 11)? "": "" );
		bp += strlen(buf);
		if( n > 3 )
			t -= t % M;     /* get rid of seconds */
	}
	if( t >= M ) {
		if( bp != buf )
			*bp++ = ' ';
		n = t / M;
		t %= M;
		sprintf(bp, "%d %s", n, (n%10 == 1 && n != 11) ? "" : "");
		bp += strlen(buf);
	}
	if( t ) {
		if( bp != buf )
			*bp++ = ' ';
		sprintf(bp, "%d %s", t, (t%10 == 1 && t != 11) ? "" : "");
	}
	return buf;
}

/*
 * Time of the day
 */
char *
tod(t)
	u_long t;
{
	struct tm *tm = localtime(&t);
	static char buf[128];

	sprintf(buf, "%d.%d.%d %02d:%02d:%02d",
		tm->tm_mday, tm->tm_mon + 1, tm->tm_year % 100,
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	return buf;
}

/*
 * Add a new entry to downtime table
 */
new_dt(dt, at, dur)
	struct DT *dt;
	u_long at;
	u_long dur;
{
	int i;

	for( i = NDT-1 ; i >= 0 ; i-- ) {
		if( dt[i].duration >= dur )
			break;
		dt[i+1] = dt[i];
	}
	dt[i+1].at = at;
	dt[i+1].duration = dur;
	dt[NDT].duration = 0;   /* keep the last element always == 0 */
}
