/*
 * Copyright (C) 1993, US Sprint
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of US Sprint not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * US SPRINT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * US SPRINT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Author: Vadim Antonov
 *
 * $Log: snmpstatd.c,v $
 * Revision 1.2  1999/07/24 15:47:06  alex
 * START debugged
 *
 * Revision 1.1.1.1  1999/07/19 18:38:32  alex
 * First revision in cvs
 *
 * Revision 1.8  1999/05/11 15:24:01  alex
 * *** empty log message ***
 *
 * Revision 1.7  1998/03/20  19:02:40  alex
 * RELCOM
 *
 * Revision 1.6  1995/12/11  14:56:22  tim
 * *** empty log message ***
 *
 * Revision 1.6  1995/12/11  14:56:22  tim
 * *** empty log message ***
 *
 * Revision 1.5  1995/08/04  14:25:02  tim
 * Added some output (avg line load etc.)
 *
 * Revision 1.5  1995/08/04  14:25:02  tim
 * Added some output (avg line load etc.)
 *
 * Revision 1.4  1994/02/04  00:54:19  avg
 * Added missing routines for HP-SUX.
 *
 * Revision 1.3  1993/09/22  23:25:53  avg
 * Got rid of chdir -- now will drop core in arbitrary places.
 *
 * Revision 1.2  1993/08/12  23:28:55  avg
 * Added "do not fork" flag.
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 * Revision 1.1  1993/08/10  17:47:57  avg
 * Initial revision
 *
 */
#ifndef lint
static char rcsid[] = "$Id: snmpstatd.c,v 1.2 1999/07/24 15:47:06 alex Exp $";
#endif

#include "manifest.h"
#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>

char *av0;
int arg_ap, arg_pp;

int hex_timestamp = 0;

int snmp_dump_packet = 0;
int debug = 0;
u_long StartTime;

int notify = -1;

void dump();
void turndebug();
void aggregate();

FILE *stf = 0;
FILE *fifstat =0;

long version;

/*
 * SNMP statistics daemon
 */
main(ac, av)
	int ac;
	char **av;
{
	extern int optind;
	extern char *optarg;
	char *cfn = NULL;
	struct GATEWAY *g;
	struct SNMP_ACTION aggr;
	char *nf = 0;
	char *sf = 0;
	char *ifs= 0;
	char *lf= 0;
	int dofork = 1;
	int check_conf = 0;

	av0 = av[0];
#ifdef DEBUG
	setbuf(stdout, 0);
	setbuf(stderr, 0);
#endif

	/*
	 * Parse the options
	 */
	for(;;) {
		switch( getopt(ac, av, "CdDFta:c:p:n:s:i:l:v:") ) {
		case EOF:
			if( optind == ac )
				break;
		default:
		usage:
			fprintf(stderr, "Usage: %s [-c conffile] [-p pollperiod] [-a aggrperiod] [-t]\n\t[-n notify_file] [-s stat_file] [-i ifsum_file] [-l lock_file]\n", av0);
			exit(1);

		case 'a':
			arg_ap = rdtime(optarg);
			if( arg_ap < 1 )
				goto usage;
			continue;

		case 'c':
			cfn = optarg;
			continue;

		case 'C':
			check_conf = 1;
			continue;

		case 'p':
			arg_pp = rdtime(optarg);
			if( arg_pp < 1 )
				goto usage;
			continue;

		case 't':
			hex_timestamp = 1;
			continue;

		case 'd':
			debug = 1;
			continue;

		case 'D':
			snmp_dump_packet = 1;
			continue;

		case 'F':
			dofork = 0;
			continue;

		case 'n':
			nf = optarg;
			continue;

		case 's':
			sf = optarg;
			continue;
 		case 'i':
			ifs = optarg;
			continue;
 		case 'l':
			lf = optarg;
			continue;
		case 'v':
			version = atoi(optarg);
			continue;
		}
		break;
	}

	/*
	 * Read in the config file
	 */
	if( rdconfig(cfn) )
		exit(1);

        /* Dump core into bas_dir */
	(void) chdir(base_dir);

	if( gwlist == 0 ) {
		fprintf(stderr, "%s: empty configuration\n", av0);
		exit(1);
	}
	if (check_conf)
		exit(0);

	/* Go into background */
	if( dofork && fork() > 0 )
		exit(0);

        if( dofork ) {
          if (setsid() == -1 ) {
            perror("setsid");
            fprintf(stderr,"setsid failed\n");
            exit(1);
          }

	/* Check the lock file */
	if( check_lock(lf) )
		exit(0);

	/* Open the notification file */
	if( nf == 0 )
		nf = notify_file;
	if( nf[0] ) {
		notify = open(nf, O_WRONLY | O_CREAT | O_APPEND, 0644);
		if( notify < 0 ) {
			fprintf(stderr, "%s: cannot open notification file %s\n", av0, nf);
			exit(1);
		}
	}

	/* Open the state file */
	if( sf == 0 )
		sf = state_file;
	if( sf[0] ) {
		stf = fopen(sf, "w");
		if( stf == 0 ) {
			fprintf(stderr, "%s: cannot create state file %s\n", av0, sf);
			exit(1);
		}
	}
        /* Open ifsum stat */
	
        if (ifs == 0 ) 
		ifs = ifsum_file;
	if( ifs[0] ) { 
		if ( ! (fifstat = fopen(ifs, "r+")) )
		       fifstat = fopen(ifs, "w+");
        	if ( fifstat == 0 ) {
			fprintf(stderr, "%s: cannot create state file %s\n", av0, ifs);
			exit(1);
		}
	}

	signal(SIGQUIT, dump);

	  if (freopen(debug_file,"w+",stderr) == 0 )
		perror("freopen");
	

          if (freopen(debug_file,"w+",stdout) == 0 )
		perror("freopen");
	
          /* To turn on/off debug messages */
	  signal(SIGUSR1, turndebug);
        }

	/* Start polling */
	for( g = gwlist ; g ; g = g->g_next )
		start_poll(g);

	/* Schedule aggregation */
	aggr.a_time = tsharp(time(0), aggr_period, 0);
	aggr.a_f = aggregate;
	timeout(&aggr);
	
	/* Start time */
	StartTime = time(0);

	/* Drop to the main loop */
	input_loop();
	/*NOTREACHED*/
}

#ifdef NORAND
/*
 * srandom and random lookalikes using old rand and srand
 */
srandom(seed)
{
	srand(seed);
}

random()
{
	return rand();
}
#endif

void turndebug() {
	debug = debug ? 0 : 1;
}
